@@ -672,6 +672,7 @@ def mk_install_tactic_cpp_internal(h_files_full_path, path):
         components.
     """
     ADD_TACTIC_DATA = []
+    ADD_SIMPLIFIER_DATA = []
     ADD_PROBE_DATA = []
     def ADD_TACTIC(name, descr, cmd):
         ADD_TACTIC_DATA.append((name, descr, cmd))
@@ -679,9 +680,13 @@ def mk_install_tactic_cpp_internal(h_files_full_path, path):
     def ADD_PROBE(name, descr, cmd):
         ADD_PROBE_DATA.append((name, descr, cmd))
 
+    def ADD_SIMPLIFIER(name, descr, cmd):
+        ADD_SIMPLIFIER_DATA.append((name, descr, cmd))
+
     eval_globals = {
         'ADD_TACTIC': ADD_TACTIC,
         'ADD_PROBE': ADD_PROBE,
+        'ADD_SIMPLIFIER': ADD_SIMPLIFIER
     }
 
     assert isinstance(h_files_full_path, list)
@@ -691,9 +696,11 @@ def mk_install_tactic_cpp_internal(h_files_full_path, path):
     fout.write('// Automatically generated file.\n')
     fout.write('#include "tactic/tactic.h"\n')
     fout.write('#include "cmd_context/tactic_cmds.h"\n')
+    fout.write('#include "cmd_context/simplifier_cmds.h"\n')
     fout.write('#include "cmd_context/cmd_context.h"\n')
     tactic_pat   = re.compile('[ \t]*ADD_TACTIC\(.*\)')
-    probe_pat    = re.compile('[ \t]*ADD_PROBE\(.*\)')   
+    probe_pat    = re.compile('[ \t]*ADD_PROBE\(.*\)')
+    simplifier_pat = re.compile('[ \t]*ADD_SIMPLIFIER\(.*\)')
     for h_file in sorted_headers_by_component(h_files_full_path):
         added_include = False
         try:
@@ -719,17 +726,31 @@ def mk_install_tactic_cpp_internal(h_files_full_path, path):
                             _logger.error("Failed processing ADD_PROBE command at '{}'\n{}".format(
                                 fullname, line))
                             raise e
+                    if simplifier_pat.match(line):
+                        if not added_include:
+                            added_include = True
+                            fout.write('#include "%s"\n' % path_after_src(h_file))
+                        try:
+                            eval(line.strip('\n '), eval_globals, None)
+                        except Exception as e:
+                            _logger.error("Failed processing ADD_SIMPLIFIER command at '{}'\n{}".format(
+                                fullname, line))
+                            raise e
+                        
         except Exception as e:
            _logger.error("Failed to read file {}\n".format(h_file))
            raise e
     # First pass will just generate the tactic factories
     fout.write('#define ADD_TACTIC_CMD(NAME, DESCR, CODE) ctx.insert(alloc(tactic_cmd, symbol(NAME), DESCR, [](ast_manager &m, const params_ref &p) { return CODE; }))\n')
     fout.write('#define ADD_PROBE(NAME, DESCR, PROBE) ctx.insert(alloc(probe_info, symbol(NAME), DESCR, PROBE))\n')
+    fout.write('#define ADD_SIMPLIFIER_CMD(NAME, DESCR, CODE) ctx.insert(alloc(simplifier_cmd, symbol(NAME), DESCR, [](auto& m, auto& p, auto &s) -> dependent_expr_simplifier* { return CODE; }))\n')
     fout.write('void install_tactics(tactic_manager & ctx) {\n')
     for data in ADD_TACTIC_DATA:
         fout.write('  ADD_TACTIC_CMD("%s", "%s", %s);\n' % data)
     for data in ADD_PROBE_DATA:
         fout.write('  ADD_PROBE("%s", "%s", %s);\n' % data)
+    for data in ADD_SIMPLIFIER_DATA:
+        fout.write('  ADD_SIMPLIFIER_CMD("%s", "%s", %s);\n' % data)
     fout.write('}\n')
     fout.close()
     return fullname
@@ -886,8 +907,7 @@ def mk_hpp_from_pyg(pyg_file, output_dir):
         hpp = os.path.join(dirname, '%s.hpp' % class_name)
         out = open(hpp, 'w')
         out.write('// Automatically generated file\n')
-        out.write('#ifndef __%s_HPP_\n' % class_name.upper())
-        out.write('#define __%s_HPP_\n' % class_name.upper())
+        out.write('#pragma once\n')
         out.write('#include "util/params.h"\n')
         if export:
             out.write('#include "util/gparams.h"\n')
@@ -919,7 +939,6 @@ def mk_hpp_from_pyg(pyg_file, output_dir):
                 out.write('  %s %s() const { return p.%s("%s", %s); }\n' %
                           (TYPE2CTYPE[param[1]], to_c_method(param[0]), TYPE2GETTER[param[1]], param[0], pyg_default_as_c_literal(param)))
         out.write('};\n')
-        out.write('#endif\n')
         out.close()
         OUTPUT_HPP_FILE.append(hpp)
 
