@@ -1253,7 +1253,9 @@ struct
   let mk_re_sort = Z3native.mk_re_sort
   let is_re_sort = Z3native.is_re_sort
   let mk_string_sort = Z3native.mk_string_sort
+  let mk_char_sort = Z3native.mk_char_sort
   let is_string_sort = Z3native.is_string_sort
+  let is_char_sort = Z3native.is_char_sort
   let mk_string = Z3native.mk_string
   let is_string = Z3native.is_string
   let get_string = Z3native.get_string
@@ -1274,6 +1276,10 @@ struct
   let mk_str_le = Z3native.mk_str_le
   let mk_str_lt = Z3native.mk_str_lt
   let mk_int_to_str = Z3native.mk_int_to_str
+  let mk_string_to_code = Z3native.mk_string_to_code
+  let mk_string_from_code = Z3native.mk_string_from_code
+  let mk_ubv_to_str = Z3native.mk_ubv_to_str
+  let mk_sbv_to_str = Z3native.mk_sbv_to_str
   let mk_seq_to_re = Z3native.mk_seq_to_re
   let mk_seq_in_re = Z3native.mk_seq_in_re
   let mk_re_plus = Z3native.mk_re_plus
@@ -1287,6 +1293,12 @@ struct
   let mk_re_complement = Z3native.mk_re_complement
   let mk_re_empty = Z3native.mk_re_empty
   let mk_re_full = Z3native.mk_re_full
+  let mk_char = Z3native.mk_char
+  let mk_char_le = Z3native.mk_char_le
+  let mk_char_to_int = Z3native.mk_char_to_int
+  let mk_char_to_bv = Z3native.mk_char_to_bv
+  let mk_char_from_bv = Z3native.mk_char_from_bv
+  let mk_char_is_digit = Z3native.mk_char_is_digit
 end
 
 module FloatingPoint =
@@ -1542,7 +1554,7 @@ struct
 
       let to_string (x:func_entry) =
         let a = get_args x in
-        let f c p = (p ^ (Expr.to_string c) ^ ", ") in
+        let f c p = ((Expr.to_string c) ^ ", " ^ p) in
         "[" ^ List.fold_right f a ((Expr.to_string (get_value x)) ^ "]")
     end
 
@@ -1734,6 +1746,39 @@ struct
   let interrupt = Z3native.interrupt
 end
 
+module Simplifier =
+struct
+  type simplifier = Z3native.simplifier
+  let gc = Z3native.context_of_simplifier
+
+  let get_help (x:simplifier) = Z3native.simplifier_get_help (gc x) x
+
+  let get_param_descrs (x:simplifier) = Z3native.simplifier_get_param_descrs (gc x) x
+
+  let get_num_simplifiers = Z3native.get_num_simplifiers
+
+  let get_simplifier_names (ctx:context) =
+    let n = get_num_simplifiers ctx in
+    let f i = Z3native.get_simplifier_name ctx i in
+    mk_list f n
+
+  let get_simplifier_description = Z3native.simplifier_get_descr
+
+  let mk_simplifier = Z3native.mk_simplifier
+
+  let and_then (ctx:context) (t1:simplifier) (t2:simplifier) (ts:simplifier list) =
+    let f p c = (match p with
+        | None -> Some c
+        | Some(x) -> Some (Z3native.simplifier_and_then ctx c x)) in
+    match (List.fold_left f None ts) with
+    | None -> Z3native.simplifier_and_then ctx t1 t2
+    | Some(x) -> let o = Z3native.simplifier_and_then ctx t2 x in
+      Z3native.simplifier_and_then ctx t1 o
+
+  let using_params = Z3native.simplifier_using_params
+  let with_ = using_params
+
+end
 
 module Statistics =
 struct
@@ -1868,6 +1913,7 @@ struct
   let mk_solver_s ctx logic = mk_solver ctx (Some (Symbol.mk_string ctx logic))
   let mk_simple_solver = Z3native.mk_simple_solver
   let mk_solver_t = Z3native.mk_solver_from_tactic
+  let add_simplifier = Z3native.solver_add_simplifier
   let translate x = Z3native.solver_translate (gc x) x
   let to_string x = Z3native.solver_to_string (gc x) x
 end
