@@ -19,6 +19,7 @@ Author:
 #include "ast/ast_ll_pp.h"
 #include "ast/ast_util.h"
 #include "ast/arith_decl_plugin.h"
+#include "ast/seq_decl_plugin.h"
 #include "ast/converters/expr_inverter.h"
 
 class basic_expr_inverter : public iexpr_inverter {
@@ -81,7 +82,7 @@ public:
      *
      */
 
-    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r, proof_ref& pr) override {
+    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r) override {
         SASSERT(f->get_family_id() == m.get_basic_family_id());
         switch (f->get_decl_kind()) {
         case OP_ITE:
@@ -233,7 +234,7 @@ public:
     }
 
 
-    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r, proof_ref& pr) override {
+    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r) override {
         SASSERT(f->get_family_id() == a.get_family_id());
         switch (f->get_decl_kind()) {
         case OP_ADD:
@@ -531,7 +532,7 @@ class bv_expr_inverter : public iexpr_inverter {
      * y := 0
      *
      */
-    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r, proof_ref& pr) override {
+    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r) override {
         SASSERT(f->get_family_id() == bv.get_family_id());
         switch (f->get_decl_kind()) {
         case OP_BADD:
@@ -611,7 +612,7 @@ public:
 
     family_id get_fid() const override { return a.get_family_id(); }
 
-    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r, proof_ref& pr) override {
+    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r) override {
         SASSERT(f->get_family_id() == a.get_family_id());
         switch (f->get_decl_kind()) {
         case OP_SELECT:
@@ -679,7 +680,7 @@ public:
      *   head(x) -> fresh
      *   x := cons(fresh, arb)
      */
-    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r, proof_ref& pr) override {
+    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r) override {
         if (dt.is_accessor(f)) {
             SASSERT(num == 1);
             if (uncnstr(args[0])) {
@@ -742,6 +743,54 @@ public:
 };
 
 
+class seq_expr_inverter : public iexpr_inverter {
+    seq_util seq;
+public:
+    seq_expr_inverter(ast_manager& m) : iexpr_inverter(m), seq(m) {}
+    
+    family_id get_fid() const override { return seq.get_family_id(); }
+
+    bool operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& r) override {
+        switch (f->get_decl_kind()) {
+        case _OP_STRING_CONCAT:
+        case OP_SEQ_CONCAT: {
+            expr* x, *y;
+            
+            if (uncnstr(args[0]) && num == 2 &&
+                args[1]->get_ref_count() == 1 && 
+                seq.str.is_concat(args[1], x, y) &&
+                uncnstr(x)) {
+                mk_fresh_uncnstr_var_for(f, r);
+                if (m_mc) {
+                    add_def(args[0], seq.str.mk_empty(args[0]->get_sort()));
+                    add_def(x, r);
+                }
+                r = seq.str.mk_concat(r, y);
+                return true;                
+            }
+                
+            if (!uncnstr(num, args)) 
+                return false;
+            mk_fresh_uncnstr_var_for(f, r);
+            if (m_mc) {
+                add_def(args[0], r);
+                for (unsigned i = 1; i < num; ++i)
+                    add_def(args[i], seq.str.mk_empty(args[0]->get_sort()));
+            }            
+            return true;
+        }
+        default:
+            return false;                
+            
+        }
+    }
+    
+    bool mk_diff(expr* t, expr_ref& r) override {
+        return false;
+    }
+    
+};
+
 
 expr_inverter::~expr_inverter() {
     for (auto* v : m_inverters)
@@ -796,10 +845,11 @@ expr_inverter::expr_inverter(ast_manager& m): iexpr_inverter(m) {
     add(alloc(array_expr_inverter, m, *this));
     add(alloc(dt_expr_inverter, m));
     add(alloc(basic_expr_inverter, m, *this));
+    add(alloc(seq_expr_inverter, m));
 }
 
 
-bool expr_inverter::operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& new_expr, proof_ref& pr) {
+bool expr_inverter::operator()(func_decl* f, unsigned num, expr* const* args, expr_ref& new_expr) {
     if (num == 0)
         return false;
             
@@ -812,7 +862,7 @@ bool expr_inverter::operator()(func_decl* f, unsigned num, expr* const* args, ex
         return false;
 
     auto* p = m_inverters.get(fid, nullptr);
-    return p && (*p)(f, num, args, new_expr, pr);       
+    return p && (*p)(f, num, args, new_expr);       
 }
 
 bool expr_inverter::mk_diff(expr* t, expr_ref& r) {
