@@ -30,14 +30,14 @@ namespace euf {
         m_nodes.push_back(n);
         m_exprs.push_back(f);
         if (is_app(f) && num_args > 0) {
-           unsigned id = to_app(f)->get_decl()->get_decl_id();
+           unsigned id = to_app(f)->get_decl()->get_small_id();
            m_decl2enodes.reserve(id+1);
            m_decl2enodes[id].push_back(n);
         }
         m_expr2enode.setx(f->get_id(), n, nullptr);
         push_node(n);
-        for (unsigned i = 0; i < num_args; ++i)
-            set_merge_enabled(args[i], true);
+        for (unsigned i = 0; i < num_args; ++i) 
+            set_cgc_enabled(args[i], true);                             
         return n;
     }
 
@@ -60,13 +60,15 @@ namespace euf {
 
 
     enode_vector const& egraph::enodes_of(func_decl* f) {
-        unsigned id = f->get_decl_id();
+        unsigned id = f->get_small_id();
         if (id < m_decl2enodes.size())
             return m_decl2enodes[id];
         return m_empty_enodes;
     }
 
     enode_bool_pair egraph::insert_table(enode* p) {
+        TRACE("euf", tout << bpp(p) << "\n");
+        //SASSERT(!m_table.contains_ptr(p));
         auto rc = m_table.insert(p);
         p->m_cg = rc.first;
         return rc;
@@ -78,9 +80,8 @@ namespace euf {
 
     void egraph::reinsert_equality(enode* p) {
         SASSERT(p->is_equality());
-        if (p->value() != l_true && p->get_arg(0)->get_root() == p->get_arg(1)->get_root()) {
-            add_literal(p, true);
-        }
+        if (p->value() != l_true && p->get_arg(0)->get_root() == p->get_arg(1)->get_root()) 
+            add_literal(p, nullptr);
     }
 
     void egraph::force_push() {
@@ -92,9 +93,7 @@ namespace euf {
             m_scopes.push_back(m_updates.size());
             m_region.push_scope();
             m_updates.push_back(update_record(m_new_th_eqs_qhead, update_record::new_th_eq_qhead()));
-            m_updates.push_back(update_record(m_new_lits_qhead, update_record::new_lits_qhead()));
         }
-        SASSERT(m_new_lits_qhead <= m_new_lits.size());
         SASSERT(m_new_th_eqs_qhead <= m_new_th_eqs.size());
     }
 
@@ -116,18 +115,16 @@ namespace euf {
             m_on_make(n);
         if (num_args == 0)             
             return n;
-        if (m.is_eq(f)) {
+        if (m.is_eq(f) && !m.is_iff(f)) {
             n->set_is_equality();
-            update_children(n);
             reinsert_equality(n);
         }
-        else {
-            auto [n2, comm] = insert_table(n);
-            if (n2 == n) 
-                update_children(n);        
-            else 
-                merge(n, n2, justification::congruence(comm));
-        }
+        auto [n2, comm] = insert_table(n);
+        if (n2 == n) 
+            update_children(n);        
+        else 
+            merge(n, n2, justification::congruence(comm, m_congruence_timestamp++));
+
         return n;
     }
 
@@ -158,11 +155,28 @@ namespace euf {
         ++m_stats.m_num_th_diseqs;
     }
 
-    void egraph::add_literal(enode* n, bool is_eq) {
-        TRACE("euf_verbose", tout << "lit: " << n->get_expr_id() << "\n";);
-        m_new_lits.push_back(enode_bool_pair(n, is_eq));
-        m_updates.push_back(update_record(update_record::new_lit()));
-        if (is_eq) ++m_stats.m_num_eqs; else ++m_stats.m_num_lits;
+    void egraph::add_literal(enode* n, enode* ante) {
+        if (!m_on_propagate_literal)
+            return;
+        if (!ante) ++m_stats.m_num_eqs; else ++m_stats.m_num_lits;
+        if (!ante)
+            m_on_propagate_literal(n, ante);
+        else if (m.is_true(ante->get_expr()) || m.is_false(ante->get_expr())) {
+            for (enode* k : enode_class(n)) {
+                if (k != ante) {
+                    //verbose_stream() << "eq: " << k->value() << " " <<ante->value() << "\n";
+                    m_on_propagate_literal(k, ante);
+                }
+            }
+        }
+        else {
+            for (enode* k : enode_class(n)) {
+                if (k->value() != ante->value()) {
+                    //verbose_stream() << "eq: " << k->value() << " " <<ante->value() << "\n";
+                    m_on_propagate_literal(k, ante);
+                }
+            }
+        }
     }
 
     void egraph::new_diseq(enode* n) {
@@ -173,7 +187,7 @@ namespace euf {
         enode* r2 = arg2->get_root();
         TRACE("euf", tout << "new-diseq:  " << bpp(r1) << " " << bpp(r2) << ": " << r1->has_th_vars() << " " << r2->has_th_vars() << "\n";);
         if (r1 == r2) {
-            add_literal(n, true);
+            add_literal(n, nullptr);
             return;
         }
         if (!r1->has_th_vars())
@@ -264,10 +278,20 @@ namespace euf {
             root->del_th_var(tid);
     }
 
-    void egraph::set_merge_enabled(enode* n, bool enable_merge) {
-        if (enable_merge != n->merge_enabled()) {
-            toggle_merge_enabled(n, false);
-            m_updates.push_back(update_record(n, update_record::toggle_merge()));
+    void egraph::set_merge_tf_enabled(enode* n, bool enable_merge_tf) {
+        if (!m.is_bool(n->get_sort()))
+            return;
+        if (enable_merge_tf != n->merge_tf()) {
+            TRACE("euf", tout << "set tf " << enable_merge_tf << " " << bpp(n) << "\n");
+            n->set_merge_tf(enable_merge_tf);
+            m_updates.push_back(update_record(n, update_record::toggle_merge_tf()));
+        }
+    }
+
+    void egraph::set_cgc_enabled(enode* n, bool enable_merge) {
+        if (enable_merge != n->cgc_enabled()) {
+            toggle_cgc_enabled(n, false);
+            m_updates.push_back(update_record(n, update_record::toggle_cgc()));
         }
     }
 
@@ -278,9 +302,9 @@ namespace euf {
         m_updates.push_back(update_record(n, update_record::set_relevant()));
     }
 
-    void egraph::toggle_merge_enabled(enode* n, bool backtracking) {
-       bool enable_merge = !n->merge_enabled();
-       n->set_merge_enabled(enable_merge);         
+    void egraph::toggle_cgc_enabled(enode* n, bool backtracking) {
+       bool enable_merge = !n->cgc_enabled();
+       n->set_cgc_enabled(enable_merge);         
        if (n->num_args() > 0) {
            if (enable_merge) {
                auto [n2, comm] = insert_table(n);
@@ -290,15 +314,18 @@ namespace euf {
            else if (n->is_cgr())
                erase_from_table(n);
        }
-       VERIFY(n->num_args() == 0 || !n->merge_enabled() || m_table.contains(n));
+       VERIFY(n->num_args() == 0 || !n->cgc_enabled() || m_table.contains(n));
     }
 
-    void egraph::set_value(enode* n, lbool value) {  
+    void egraph::set_value(enode* n, lbool value, justification j) {  
         if (n->value() == l_undef) {
             force_push();
             TRACE("euf", tout << bpp(n) << " := " << value << "\n";);
             n->set_value(value);
+            n->m_lit_justification = j;
             m_updates.push_back(update_record(n, update_record::value_assignment()));
+            if (n->is_equality() && n->value() == l_false) 
+                new_diseq(n);
         }
     }
 
@@ -328,7 +355,6 @@ namespace euf {
         num_scopes -= m_num_scopes;
         m_num_scopes = 0;
 
-        SASSERT(m_new_lits_qhead <= m_new_lits.size());
         unsigned old_lim = m_scopes.size() - num_scopes;
         unsigned num_updates = m_scopes[old_lim];
         auto undo_node = [&]() {
@@ -340,7 +366,7 @@ namespace euf {
             m_expr2enode[e->get_id()] = nullptr;
             n->~enode();
             if (is_app(e) && n->num_args() > 0)
-               m_decl2enodes[to_app(e)->get_decl()->get_decl_id()].pop_back();
+               m_decl2enodes[to_app(e)->get_decl()->get_small_id()].pop_back();
             m_nodes.pop_back();
             m_exprs.pop_back();
         };
@@ -351,8 +377,11 @@ namespace euf {
             case update_record::tag_t::is_add_node:
                 undo_node();
                 break;
-            case update_record::tag_t::is_toggle_merge:
-                toggle_merge_enabled(p.r1, true);
+            case update_record::tag_t::is_toggle_cgc:
+                toggle_cgc_enabled(p.r1, true);
+                break;
+            case update_record::tag_t::is_toggle_merge_tf:
+                p.r1->set_merge_tf(!p.r1->merge_tf());
                 break;
             case update_record::tag_t::is_set_parent:
                 undo_eq(p.r1, p.n1, p.r2_num_parents);
@@ -364,18 +393,12 @@ namespace euf {
                 SASSERT(p.r1->get_th_var(p.m_th_id) != null_theory_var);
                 p.r1->replace_th_var(p.m_old_th_var, p.m_th_id);
                 break;
-            case update_record::tag_t::is_new_lit:
-                m_new_lits.pop_back();
-                break;
             case update_record::tag_t::is_new_th_eq:
                 m_new_th_eqs.pop_back();
                 break;
             case update_record::tag_t::is_new_th_eq_qhead:
                 m_new_th_eqs_qhead = p.qhead;
                 break;
-            case update_record::tag_t::is_new_lits_qhead:
-                m_new_lits_qhead = p.qhead;
-                break;
             case update_record::tag_t::is_inconsistent:
                 m_inconsistent = p.m_inconsistent;
                 break;
@@ -410,7 +433,6 @@ namespace euf {
         m_region.pop_scope(num_scopes);  
         m_to_merge.reset();
 
-        SASSERT(m_new_lits_qhead <= m_new_lits.size());
         SASSERT(m_new_th_eqs_qhead <= m_new_th_eqs.size());
 
         // DEBUG_CODE(invariant(););
@@ -418,7 +440,7 @@ namespace euf {
 
     void egraph::merge(enode* n1, enode* n2, justification j) {
 
-        if (!n1->merge_enabled() && !n2->merge_enabled())
+        if (!n1->cgc_enabled() && !n2->cgc_enabled())
             return;
         SASSERT(n1->get_sort() == n2->get_sort());
         enode* r1 = n1->get_root();
@@ -435,6 +457,7 @@ namespace euf {
             set_conflict(n1, n2, j);
             return;
         }
+        
         if (r1->value() != r2->value() && r1->value() != l_undef && r2->value() != l_undef) {
             SASSERT(m.is_bool(r1->get_expr()));
             set_conflict(n1, n2, j);
@@ -446,11 +469,7 @@ namespace euf {
             std::swap(n1, n2);
         }
 
-        if (j.is_congruence() && (m.is_false(r2->get_expr()) || m.is_true(r2->get_expr())))
-            add_literal(n1, false);
-        if (n1->is_equality() && n1->value() == l_false)
-            new_diseq(n1);
-        remove_parents(r1, r2);
+        remove_parents(r1);
         push_eq(r1, n1, r2->num_parents());
         merge_justification(n1, n2, j);
         for (enode* c : enode_class(n1)) 
@@ -459,20 +478,29 @@ namespace euf {
         r2->inc_class_size(r1->class_size());   
         merge_th_eq(r1, r2);
         reinsert_parents(r1, r2);
+        if (j.is_congruence() && (m.is_false(r2->get_expr()) || m.is_true(r2->get_expr())))
+            add_literal(n1, r2);
+        else if (n2->value() != l_undef && n1->value() != n2->value()) 
+            add_literal(n1, n2);
+        else if (n1->value() != l_undef && n2->value() != n1->value()) 
+            add_literal(n2, n1);
+
         for (auto& cb : m_on_merge)
             cb(r2, r1);
     }
 
-    void egraph::remove_parents(enode* r1, enode* r2) {
-        for (enode* p : enode_parents(r1)) {
+    void egraph::remove_parents(enode* r) {
+        TRACE("euf", tout << bpp(r) << "\n");
+        for (enode* p : enode_parents(r)) {
             if (p->is_marked1())
                 continue;
-            if (p->merge_enabled()) {
+            if (p->cgc_enabled()) {
                 if (!p->is_cgr())
                     continue;
                 SASSERT(m_table.contains_ptr(p));
                 p->mark1();
                 erase_from_table(p);
+                CTRACE("euf", m_table.contains_ptr(p), tout << bpp(p) << "\n"; display(tout));
                 SASSERT(!m_table.contains_ptr(p));
             }
             else if (p->is_equality())
@@ -485,8 +513,8 @@ namespace euf {
             if (!p->is_marked1())
                 continue;
             p->unmark1();
-            TRACE("euf", tout << "reinsert " << bpp(r1) << " " << bpp(r2) << " " << bpp(p) << " " << p->merge_enabled() << "\n";);
-            if (p->merge_enabled()) {
+            TRACE("euf", tout << "reinsert " << bpp(r1) << " " << bpp(r2) << " " << bpp(p) << " " << p->cgc_enabled() << "\n";);
+            if (p->cgc_enabled()) {
                 auto [p_other, comm] = insert_table(p);
                 SASSERT(m_table.contains_ptr(p) == (p_other == p));
                 TRACE("euf", tout << "other " << bpp(p_other) << "\n";);
@@ -530,9 +558,9 @@ namespace euf {
         for (auto it = begin; it != end; ++it) {
             enode* p = *it;
             TRACE("euf", tout << "erase " << bpp(p) << "\n";);
-            SASSERT(!p->merge_enabled() || m_table.contains_ptr(p));
-            SASSERT(!p->merge_enabled() || p->is_cgr());
-            if (p->merge_enabled())
+            SASSERT(!p->cgc_enabled() || m_table.contains_ptr(p));
+            SASSERT(!p->cgc_enabled() || p->is_cgr());
+            if (p->cgc_enabled())
                 erase_from_table(p);
         }
 
@@ -540,7 +568,7 @@ namespace euf {
             c->m_root = r1;
 
         for (enode* p : enode_parents(r1)) 
-            if (p->merge_enabled() && (p->is_cgr() || !p->congruent(p->m_cg))) 
+            if (p->cgc_enabled() && (p->is_cgr() || !p->congruent(p->m_cg))) 
                 insert_table(p);                    
         r2->m_parents.shrink(r2_num_parents);
         unmerge_justification(n1);
@@ -548,16 +576,14 @@ namespace euf {
 
 
     bool egraph::propagate() {
-        SASSERT(m_new_lits_qhead <= m_new_lits.size());
         SASSERT(m_num_scopes == 0 || m_to_merge.empty());
         force_push();
         for (unsigned i = 0; i < m_to_merge.size() && m.limit().inc() && !inconsistent(); ++i) {
             auto const& w = m_to_merge[i];
-            merge(w.a, w.b, justification::congruence(w.commutativity));                
+            merge(w.a, w.b, justification::congruence(w.commutativity, m_congruence_timestamp++));                
         }
         m_to_merge.reset();
         return 
-            (m_new_lits_qhead < m_new_lits.size()) || 
             (m_new_th_eqs_qhead < m_new_th_eqs.size()) ||
             inconsistent();
     }
@@ -570,6 +596,7 @@ namespace euf {
         m_updates.push_back(update_record(false, update_record::inconsistent()));
         m_n1 = n1;
         m_n2 = n2;
+        TRACE("euf", tout << "conflict " << bpp(n1) << " " << bpp(n2) << " " << j << "\n");
         m_justification = j;
     }
 
@@ -648,7 +675,7 @@ namespace euf {
         SASSERT(n1->get_decl() == n2->get_decl());
         m_uses_congruence = true;
         if (m_used_cc && !comm) { 
-            m_used_cc(to_app(n1->get_expr()), to_app(n2->get_expr()));
+            m_used_cc(n1->get_app(), n2->get_app());
         }
         if (comm && 
             n1->get_arg(0)->get_root() == n2->get_arg(1)->get_root() &&
@@ -657,6 +684,7 @@ namespace euf {
             push_lca(n1->get_arg(1), n2->get_arg(0));
             return;
         }
+        TRACE("euf_verbose", tout << bpp(n1) << " " << bpp(n2) << "\n");
             
         for (unsigned i = 0; i < n1->num_args(); ++i) 
             push_lca(n1->get_arg(i), n2->get_arg(i));
@@ -705,16 +733,28 @@ namespace euf {
     }
 
     template <typename T>
-    void egraph::explain(ptr_vector<T>& justifications) {
+    void egraph::explain(ptr_vector<T>& justifications, cc_justification* cc) {
         SASSERT(m_inconsistent);
         push_todo(m_n1);
         push_todo(m_n2);
-        explain_eq(justifications, m_n1, m_n2, m_justification);
-        explain_todo(justifications);
+        explain_eq(justifications, cc, m_n1, m_n2, m_justification);
+        explain_todo(justifications, cc);
     }
 
     template <typename T>
-    void egraph::explain_eq(ptr_vector<T>& justifications, enode* a, enode* b) {
+    void egraph::explain_eq(ptr_vector<T>& justifications, cc_justification* cc, enode* a, enode* b, justification const& j) {
+        TRACE("euf_verbose", tout << "explain-eq: " << bpp(a) << " == " << bpp(b) << " jst: " << j << "\n";);
+        if (j.is_external())
+            justifications.push_back(j.ext<T>());
+        else if (j.is_congruence()) 
+            push_congruence(a, b, j.is_commutative());
+        if (cc && j.is_congruence()) 
+            cc->push_back(std::tuple(a->get_app(), b->get_app(), j.timestamp(), j.is_commutative()));
+    }
+
+
+    template <typename T>
+    void egraph::explain_eq(ptr_vector<T>& justifications, cc_justification* cc, enode* a, enode* b) {
         SASSERT(a->get_root() == b->get_root());
         
         enode* lca = find_lca(a, b);
@@ -723,33 +763,43 @@ namespace euf {
         push_to_lca(b, lca);
         if (m_used_eq)
             m_used_eq(a->get_expr(), b->get_expr(), lca->get_expr());
-        explain_todo(justifications);
+        explain_todo(justifications, cc);
     }
 
     template <typename T>
-    unsigned egraph::explain_diseq(ptr_vector<T>& justifications, enode* a, enode* b) {
+    unsigned egraph::explain_diseq(ptr_vector<T>& justifications, cc_justification* cc, enode* a, enode* b) {
         enode* ra = a->get_root(), * rb = b->get_root();
         SASSERT(ra != rb);
         if (ra->interpreted() && rb->interpreted()) {
-            explain_eq(justifications, a, ra);
-            explain_eq(justifications, b, rb);
+            explain_eq(justifications, cc, a, ra);
+            explain_eq(justifications, cc, b, rb);
             return sat::null_bool_var;
         }
         enode* r = tmp_eq(ra, rb);
         SASSERT(r && r->get_root()->value() == l_false);
-        explain_eq(justifications, r, r->get_root());
+        explain_eq(justifications, cc, r, r->get_root());
         return r->get_root()->bool_var();
     }
 
 
     template <typename T>
-    void egraph::explain_todo(ptr_vector<T>& justifications) {
+    void egraph::explain_todo(ptr_vector<T>& justifications, cc_justification* cc) {
         for (unsigned i = 0; i < m_todo.size(); ++i) {
             enode* n = m_todo[i];
-            if (n->m_target && !n->is_marked1()) {
+            if (n->is_marked1())
+                continue;
+            if (n->m_target) {
                 n->mark1();
                 CTRACE("euf_verbose", m_display_justification, n->m_justification.display(tout << n->get_expr_id() << " = " << n->m_target->get_expr_id() << " ", m_display_justification) << "\n";);
-                explain_eq(justifications, n, n->m_target, n->m_justification);
+                explain_eq(justifications, cc, n, n->m_target, n->m_justification);
+            }
+            else if (!n->is_marked1() && n->value() != l_undef) {
+                n->mark1();
+                if (m.is_true(n->get_expr()) || m.is_false(n->get_expr()))
+                    continue;
+                justification j = n->m_lit_justification;
+                SASSERT(j.is_external());
+                justifications.push_back(j.ext<T>());
             }
         }
     }
@@ -758,7 +808,7 @@ namespace euf {
         for (enode* n : m_nodes)
             n->invariant(*this);
         for (enode* n : m_nodes)
-            if (n->merge_enabled() && n->num_args() > 0 && (!m_table.find(n) || n->get_root() != m_table.find(n)->get_root())) {
+            if (n->cgc_enabled() && n->num_args() > 0 && (!m_table.find(n) || n->get_root() != m_table.find(n)->get_root())) {
                 CTRACE("euf", !m_table.find(n), tout << "node is not in table\n";);
                 CTRACE("euf", m_table.find(n), tout << "root " << bpp(n->get_root()) << " table root " << bpp(m_table.find(n)->get_root()) << "\n";);
                 TRACE("euf", display(tout << bpp(n) << " is not closed under congruence\n"););
@@ -793,7 +843,7 @@ namespace euf {
             }
         };
         if (n->bool_var() != sat::null_bool_var) 
-            out << "[b" << n->bool_var() << " := " << value_of() << (n->merge_tf() ? "" : " no merge") << "] ";
+            out << "[b" << n->bool_var() << " := " << value_of() << (n->cgc_enabled() ? "" : " no-cgc") << (n->merge_tf()? " merge-tf" : "") << "] ";
         if (n->has_th_vars()) {
             out << "[t";
             for (auto const& v : enode_th_vars(n))
@@ -810,7 +860,6 @@ namespace euf {
 
     std::ostream& egraph::display(std::ostream& out) const {
         out << "updates " << m_updates.size() << "\n";
-        out << "newlits " << m_new_lits.size()   << " qhead: " << m_new_lits_qhead << "\n";
         out << "neweqs  " << m_new_th_eqs.size() << " qhead: " << m_new_th_eqs_qhead << "\n";
         m_table.display(out);
         unsigned max_args = 0;
@@ -848,7 +897,8 @@ namespace euf {
             n2->set_value(n1->value());
             n2->m_bool_var = n1->m_bool_var;
             n2->m_commutative = n1->m_commutative;
-            n2->m_merge_enabled = n1->m_merge_enabled;
+            n2->m_cgc_enabled = n1->m_cgc_enabled;
+            n2->m_merge_tf_enabled = n1->m_merge_tf_enabled;
             n2->m_is_equality = n1->m_is_equality;            
         }
         for (unsigned i = 0; i < src.m_nodes.size(); ++i) {             
@@ -869,16 +919,20 @@ namespace euf {
     }
 }
 
-template void euf::egraph::explain(ptr_vector<int>& justifications);
-template void euf::egraph::explain_todo(ptr_vector<int>& justifications);
-template void euf::egraph::explain_eq(ptr_vector<int>& justifications, enode* a, enode* b);
-template unsigned euf::egraph::explain_diseq(ptr_vector<int>& justifications, enode* a, enode* b);
+template void euf::egraph::explain(ptr_vector<int>& justifications, cc_justification*);
+template void euf::egraph::explain_todo(ptr_vector<int>& justifications, cc_justification*);
+template void euf::egraph::explain_eq(ptr_vector<int>& justifications, cc_justification*, enode* a, enode* b);
+template unsigned euf::egraph::explain_diseq(ptr_vector<int>& justifications, cc_justification*, enode* a, enode* b);
 
-template void euf::egraph::explain(ptr_vector<size_t>& justifications);
-template void euf::egraph::explain_todo(ptr_vector<size_t>& justifications);
-template void euf::egraph::explain_eq(ptr_vector<size_t>& justifications, enode* a, enode* b);
-template unsigned euf::egraph::explain_diseq(ptr_vector<size_t>& justifications, enode* a, enode* b);
+template void euf::egraph::explain(ptr_vector<size_t>& justifications, cc_justification*);
+template void euf::egraph::explain_todo(ptr_vector<size_t>& justifications, cc_justification*);
+template void euf::egraph::explain_eq(ptr_vector<size_t>& justifications, cc_justification*, enode* a, enode* b);
+template unsigned euf::egraph::explain_diseq(ptr_vector<size_t>& justifications, cc_justification*, enode* a, enode* b);
 
+template void euf::egraph::explain(ptr_vector<expr_dependency>& justifications, cc_justification*);
+template void euf::egraph::explain_todo(ptr_vector<expr_dependency>& justifications, cc_justification*);
+template void euf::egraph::explain_eq(ptr_vector<expr_dependency>& justifications, cc_justification*, enode* a, enode* b);
+template unsigned euf::egraph::explain_diseq(ptr_vector<expr_dependency>& justifications, cc_justification*, enode* a, enode* b);
 
 
 #if 0
