@@ -33,7 +33,7 @@ Notes:
 #include "ast/datatype_decl_plugin.h"
 #include "ast/recfun_decl_plugin.h"
 #include "ast/rewriter/seq_rewriter.h"
-#include "tactic/generic_model_converter.h"
+#include "ast/converters/generic_model_converter.h"
 #include "solver/solver.h"
 #include "solver/check_logic.h"
 #include "solver/progress_callback.h"
@@ -90,13 +90,26 @@ public:
     vector<macro_decl>::iterator end() const { return m_decls->end(); }
 };
 
+
+class proof_cmds {
+public:
+    virtual ~proof_cmds() {}
+    virtual void add_literal(expr* e) = 0;
+    virtual void end_assumption() = 0;
+    virtual void end_infer() = 0;
+    virtual void end_deleted() = 0;
+    virtual void updt_params(params_ref const& p) = 0;
+    virtual void register_on_clause(void* ctx, user_propagator::on_clause_eh_t& on_clause) = 0;
+};
+
+
 /**
    \brief Generic wrapper.
 */
 class object_ref {
     unsigned m_ref_count = 0;
 public:
-    virtual ~object_ref() {}
+    virtual ~object_ref() = default;
     virtual void finalize(cmd_context & ctx) = 0;
     void inc_ref(cmd_context & ctx) {
         m_ref_count++;
@@ -148,6 +161,7 @@ struct builtin_decl {
 
 class opt_wrapper : public check_sat_result {
 public:
+    opt_wrapper(ast_manager& m): check_sat_result(m) {}
     virtual bool empty() = 0;
     virtual void push() = 0;
     virtual void pop(unsigned n) = 0;
@@ -172,6 +186,7 @@ public:
     bool owns_manager() const { return m_manager != nullptr; }
 };
 
+
 class cmd_context : public progress_callback, public tactic_manager, public ast_printer_context {
 public:
     enum status {
@@ -191,24 +206,41 @@ public:
         ~scoped_watch() { m_ctx.m_watch.stop(); }
     };
 
+    struct scoped_redirect {
+        cmd_context& m_ctx;
+        std::ostream& m_verbose;
+        std::ostream* m_warning;
+
+        scoped_redirect(cmd_context& ctx): m_ctx(ctx), m_verbose(verbose_stream()), m_warning(warning_stream()) {
+            set_warning_stream(&(*m_ctx.m_diagnostic));
+            set_verbose_stream(m_ctx.diagnostic_stream());
+        }
+
+        ~scoped_redirect() {
+            set_verbose_stream(m_verbose);
+            set_warning_stream(m_warning);
+        }
+    };
+
     
 
 protected:
-    ast_context_params                  m_params;
+    ast_context_params           m_params;
     bool                         m_main_ctx;
     symbol                       m_logic;
-    bool                         m_interactive_mode;
-    bool                         m_global_decls;
+    bool                         m_interactive_mode = false;
+    bool                         m_global_decls = false;
     bool                         m_print_success;
-    unsigned                     m_random_seed;
-    bool                         m_produce_unsat_cores;
-    bool                         m_produce_unsat_assumptions;
-    bool                         m_produce_assignments;
-    status                       m_status;
-    bool                         m_numeral_as_real;
-    bool                         m_ignore_check;      // used by the API to disable check-sat() commands when parsing SMT 2.0 files.
-    bool                         m_exit_on_error;
-    bool                         m_allow_duplicate_declarations { false };
+    unsigned                     m_random_seed = 0;
+    bool                         m_produce_unsat_cores = false;
+    bool                         m_produce_unsat_assumptions = false;
+    bool                         m_produce_assignments = false;
+    status                       m_status = UNKNOWN;
+    bool                         m_numeral_as_real = false;
+    bool                         m_ignore_check = false;      // used by the API to disable check-sat() commands when parsing SMT 2.0 files.
+    bool                         m_exit_on_error = false;
+    bool                         m_allow_duplicate_declarations = false;
+    scoped_ptr<proof_cmds>       m_proof_cmds;
 
     static std::ostringstream    g_error_stream;
 
@@ -216,9 +248,9 @@ protected:
     sref_vector<generic_model_converter> m_mcs;
     ast_manager *                m_manager;
     bool                         m_own_manager;
-    bool                         m_manager_initialized;
-    pdecl_manager *              m_pmanager;
-    sexpr_manager *              m_sexpr_manager;
+    bool                         m_manager_initialized = false;
+    pdecl_manager *              m_pmanager = nullptr;
+    sexpr_manager *              m_sexpr_manager = nullptr;
     check_logic                  m_check_logic;
     stream_ref                   m_regular;
     stream_ref                   m_diagnostic;
@@ -267,6 +299,7 @@ protected:
         cmd_context &             m_owner;
         datatype_util             m_dt_util;
     public:
+        void reset() { m_dt_util.reset(); }
         dt_eh(cmd_context & owner);
         ~dt_eh() override;
         void operator()(sort * dt, pdecl* pd) override;
@@ -285,7 +318,7 @@ protected:
     void register_builtin_ops(decl_plugin * p);
     void load_plugin(symbol const & name, bool install_names, svector<family_id>& fids);
     void init_manager_core(bool new_manager);
-    void init_manager();
+
     void init_external_manager();
     void reset_cmds();
     void finalize_cmds();
@@ -361,6 +394,7 @@ public:
     bool produce_unsat_cores() const;
     bool well_sorted_check_enabled() const;
     bool validate_model_enabled() const;
+    bool has_assertions() const { return !m_assertions.empty(); }
     void set_produce_models(bool flag);
     void set_produce_unsat_cores(bool flag);
     void set_produce_proofs(bool flag);
@@ -379,6 +413,12 @@ public:
     pdecl_manager & pm() const { if (!m_pmanager) const_cast<cmd_context*>(this)->init_manager(); return *m_pmanager; }
     sexpr_manager & sm() const { if (!m_sexpr_manager) const_cast<cmd_context*>(this)->m_sexpr_manager = alloc(sexpr_manager); return *m_sexpr_manager; }
 
+    proof_cmds* get_proof_cmds() { return m_proof_cmds.get(); }
+    void init_manager();
+    solver* get_solver() { return m_solver.get(); }
+    void set_solver(solver* s) { m_solver = s; }
+    void set_proof_cmds(proof_cmds* pc) { m_proof_cmds = pc; }
+
     void set_solver_factory(solver_factory * s);
     void set_check_sat_result(check_sat_result * r) { m_check_sat_result = r; }
     check_sat_result * get_check_sat_result() const { return m_check_sat_result.get(); }
@@ -483,7 +523,8 @@ public:
 
     ptr_vector<expr> const& assertions() const { return m_assertions; }
     ptr_vector<expr> const& assertion_names() const { return m_assertion_names; }
-    expr_ref_vector tracked_assertions();
+    vector<std::pair<expr*,expr*>> tracked_assertions();
+    void reset_tracked_assertions();
 
     /**
        \brief Hack: consume assertions if there are no scopes.
