@@ -31,14 +31,12 @@
 #include "math/lp/lar_constraints.h"
 #include "math/lp/lar_core_solver.h"
 #include "math/lp/numeric_pair.h"
-#include "math/lp/scaler.h"
 #include "math/lp/lp_primal_core_solver.h"
 #include "math/lp/random_updater.h"
 #include "util/stacked_value.h"
 #include "math/lp/stacked_vector.h"
 #include "math/lp/implied_bound.h"
 #include "math/lp/bound_analyzer_on_row.h"
-#include "math/lp/conversion_helper.h"
 #include "math/lp/int_solver.h"
 #include "math/lp/nra_solver.h"
 #include "math/lp/lp_types.h"
@@ -93,7 +91,7 @@ class lar_solver : public column_namer {
     unsigned_vector                                     m_row_bounds_to_replay;
     
     u_set                                               m_basic_columns_with_changed_cost;
-    // these are basic columns with the value changed, so the the corresponding row in the tableau
+    // these are basic columns with the value changed, so the corresponding row in the tableau
     // does not sum to zero anymore
     u_set                                               m_incorrect_columns;
     // copy of m_r_solver.inf_set()
@@ -113,8 +111,6 @@ class lar_solver : public column_namer {
     // end of fields
 
     ////////////////// methods ////////////////////////////////
-    static_matrix<double, double> & A_d();
-    static_matrix<double, double > const & A_d() const;
     
     static bool valid_index(unsigned j) { return static_cast<int>(j) >= 0;}
     const lar_term & get_term(unsigned j) const;
@@ -125,7 +121,6 @@ class lar_solver : public column_namer {
     bool term_is_int(const lar_term * t) const;
     bool term_is_int(const vector<std::pair<mpq, unsigned int>> & coeffs) const;
     void add_non_basic_var_to_core_fields(unsigned ext_j, bool is_int);
-    void add_new_var_to_core_fields_for_doubles(bool register_in_basis);
     void add_new_var_to_core_fields_for_mpq(bool register_in_basis);
     mpq adjust_bound_for_int(lpvar j, lconstraint_kind&, const mpq&);
 
@@ -134,7 +129,6 @@ class lar_solver : public column_namer {
     var_index add_term_undecided(const vector<std::pair<mpq, var_index>> & coeffs);
     bool term_coeffs_are_ok(const vector<std::pair<mpq, var_index>> & coeffs);
     void push_term(lar_term* t);
-    void add_row_for_term(const lar_term * term, unsigned term_ext_index);
     void add_row_from_term_no_constraint(const lar_term * term, unsigned term_ext_index);
     void add_basic_var_to_core_fields();
     bool compare_values(impq const& lhs, lconstraint_kind k, const mpq & rhs);
@@ -163,39 +157,28 @@ class lar_solver : public column_namer {
     unsigned row_of_basic_column(unsigned) const;
     void decide_on_strategy_and_adjust_initial_state();
     void adjust_initial_state();
-    void adjust_initial_state_for_lu();
     void adjust_initial_state_for_tableau_rows();
-    void fill_last_row_of_A_d(static_matrix<double, double> & A, const lar_term* ls);
-    bool use_lu() const;
     bool sizes_are_correct() const;
     bool implied_bound_is_correctly_explained(implied_bound const & be, const vector<std::pair<mpq, unsigned>> & explanation) const;
     
-    template <typename T>
-    void analyze_new_bounds_on_row_tableau(
-        unsigned row_index,
-        lp_bound_propagator<T> & bp ) {
-        
-        if (A_r().m_rows[row_index].size() > settings().max_row_length_for_bound_propagation
-            || row_has_a_big_num(row_index))
-            return;
-        lp_assert(use_tableau());
-        
-        bound_analyzer_on_row<row_strip<mpq>, lp_bound_propagator<T>>::analyze_row(A_r().m_rows[row_index],
-                                                                                   null_ci,
-                                                                                   zero_of_type<numeric_pair<mpq>>(),
-                                                                                   row_index,
-                                                                                   bp
-                                                                                   );
-    }
 
     void substitute_basis_var_in_terms_for_row(unsigned i);
+    
     template <typename T>
-    void calculate_implied_bounds_for_row(unsigned i, lp_bound_propagator<T> & bp) {
-        SASSERT(use_tableau());
-        analyze_new_bounds_on_row_tableau(i, bp);
+    unsigned calculate_implied_bounds_for_row(unsigned row_index, lp_bound_propagator<T> & bp) {
+        
+        if (A_r().m_rows[row_index].size() > settings().max_row_length_for_bound_propagation  || row_has_a_big_num(row_index))
+            return 0;
+        
+        return bound_analyzer_on_row<row_strip<mpq>, lp_bound_propagator<T>>::analyze_row(
+            A_r().m_rows[row_index],
+            null_ci,
+            zero_of_type<numeric_pair<mpq>>(),
+            row_index,
+            bp);
     }
+    
     static void clean_popped_elements(unsigned n, u_set& set);
-    static void shrink_inf_set_after_pop(unsigned n, u_set & set);
     bool maximize_term_on_tableau(const lar_term & term,
                                   impq &term_max);
     bool costs_are_zeros_for_r_solver() const;
@@ -209,12 +192,9 @@ class lar_solver : public column_namer {
     void set_lower_bound_witness(var_index j, constraint_index ci);
     void substitute_terms_in_linear_expression( const vector<std::pair<mpq, var_index>>& left_side_with_terms,
                                                 vector<std::pair<mpq, var_index>> &left_side) const;
-    void detect_rows_of_bound_change_column_for_nbasic_column(unsigned j);
+    
     void detect_rows_of_bound_change_column_for_nbasic_column_tableau(unsigned j);
-    bool use_tableau() const;
     bool use_tableau_costs() const;
-    void detect_rows_of_column_with_bound_change(unsigned j);
-    void adjust_x_of_column(unsigned j);
     bool tableau_with_costs() const;
     bool costs_are_used() const;
     void change_basic_columns_dependend_on_a_given_nb_column(unsigned j, const numeric_pair<mpq> & delta);
@@ -223,28 +203,16 @@ class lar_solver : public column_namer {
     void insert_row_with_changed_bounds(unsigned rid);
     void detect_rows_with_changed_bounds_for_column(unsigned j);
     void detect_rows_with_changed_bounds();
-    void set_value_for_nbasic_column(unsigned j, const impq & new_val);
-    void update_x_and_inf_costs_for_columns_with_changed_bounds();
+
     void update_x_and_inf_costs_for_columns_with_changed_bounds_tableau();
     void solve_with_core_solver();
     numeric_pair<mpq> get_basic_var_value_from_row(unsigned i);
-    template <typename K, typename L>
-    void add_last_rows_to_lu(lp_primal_core_solver<K,L> & s);
-    bool x_is_correct() const;
-    void fill_last_row_of_A_r(static_matrix<mpq, numeric_pair<mpq>> & A, const lar_term * ls);
-    template <typename U, typename V>
-    void create_matrix_A(static_matrix<U, V> & matr);
-    template <typename U, typename V>
-    void copy_from_mpq_matrix(static_matrix<U, V> & matr);
-    bool try_to_set_fixed(column_info<mpq> & ci);
     bool all_constrained_variables_are_registered(const vector<std::pair<mpq, var_index>>& left_side);
     bool all_constraints_hold() const;
     bool constraint_holds(const lar_base_constraint & constr, std::unordered_map<var_index, mpq> & var_map) const;
-    bool the_relations_are_of_same_type(const vector<std::pair<mpq, unsigned>> & evidence, lconstraint_kind & the_kind_of_sum) const;
     static void register_in_map(std::unordered_map<var_index, mpq> & coeffs, const lar_base_constraint & cn, const mpq & a);
     static void register_monoid_in_map(std::unordered_map<var_index, mpq> & coeffs, const mpq & a, unsigned j);
     bool the_left_sides_sum_to_zero(const vector<std::pair<mpq, unsigned>> & evidence) const;
-    bool the_right_sides_do_not_sum_to_zero(const vector<std::pair<mpq, unsigned>> & evidence);
     bool explanation_is_correct(explanation&) const;
     bool inf_explanation_is_correct() const;
     mpq sum_of_right_sides_of_explanation(explanation &) const;
@@ -254,7 +222,6 @@ class lar_solver : public column_namer {
         int inf_sign) const;
     mpq get_left_side_val(const lar_base_constraint &  cns, const std::unordered_map<var_index, mpq> & var_map) const;
     void fill_var_set_for_random_update(unsigned sz, var_index const * vars, vector<unsigned>& column_list);
-    void pivot_fixed_vars_from_basis();
     bool column_represents_row_in_tableau(unsigned j);
     void make_sure_that_the_bottom_right_elem_not_zero_in_tableau(unsigned i, unsigned j);
     void remove_last_row_and_column_from_tableau(unsigned j);
@@ -264,30 +231,16 @@ class lar_solver : public column_namer {
     void remove_last_column_from_tableau();
     void pop_tableau();
     void clean_inf_set_of_r_solver_after_pop();
-    void shrink_explanation_to_minimum(vector<std::pair<mpq, constraint_index>> & explanation) const;
     inline bool column_value_is_integer(unsigned j) const { return get_column_value(j).is_int(); }
     bool model_is_int_feasible() const;
-    inline
-    indexed_vector<mpq> & get_column_in_lu_mode(unsigned j) {
-        m_column_buffer.clear();
-        m_column_buffer.resize(A_r().row_count());
-        m_mpq_lar_core_solver.m_r_solver.solve_Bd(j, m_column_buffer);
-        return m_column_buffer;
-    }
+    
     bool bound_is_integer_for_integer_column(unsigned j, const mpq & right_side) const;
-    inline unsigned get_base_column_in_row(unsigned row_index) const {
-        return m_mpq_lar_core_solver.m_r_solver.get_base_column_in_row(row_index);
-    }
     inline lar_core_solver & get_core_solver() { return m_mpq_lar_core_solver; }
-    void catch_up_in_updating_int_solver();
     var_index to_column(unsigned ext_j) const;
     void fix_terms_with_rounded_columns();
-    void update_delta_for_terms(const impq & delta, unsigned j, const vector<unsigned>&);
-    void fill_vars_to_terms(vector<vector<unsigned>> & vars_to_terms);
     bool remove_from_basis(unsigned);
     lar_term get_term_to_maximize(unsigned ext_j) const;
     bool sum_first_coords(const lar_term& t, mpq & val) const;
-    void collect_rounded_rows_to_fix();
     void register_normalized_term(const lar_term&, lpvar);
     void deregister_normalized_term(const lar_term&);
 
@@ -298,63 +251,81 @@ class lar_solver : public column_namer {
 public:
     // this function just looks at the status
     bool is_feasible() const;
+
     const map<mpq, unsigned, obj_hash<mpq>, default_eq<mpq>>& fixed_var_table_int() const {
         return m_fixed_var_table_int;
     }
-    map<mpq, unsigned, obj_hash<mpq>, default_eq<mpq>>& fixed_var_table_int() {
-        return m_fixed_var_table_int;
-    }
+
+    
     const map<mpq, unsigned, obj_hash<mpq>, default_eq<mpq>>& fixed_var_table_real() const {
         return m_fixed_var_table_real;
     }
+
     map<mpq, unsigned, obj_hash<mpq>, default_eq<mpq>>& fixed_var_table_real() {
         return m_fixed_var_table_real;
     }
 
     bool find_in_fixed_tables(const rational& mpq, bool is_int, unsigned & j) const {
-        return is_int? fixed_var_table_int().find(mpq, j) :
-            fixed_var_table_real().find(mpq, j);
+        return is_int? fixed_var_table_int().find(mpq, j) : fixed_var_table_real().find(mpq, j);
     }
     
     template <typename T> void remove_non_fixed_from_table(T&);
+
     unsigned external_to_column_index(unsigned) const;
+
     bool inside_bounds(lpvar, const impq&) const;
+
     inline void set_column_value(unsigned j, const impq& v) {
         m_mpq_lar_core_solver.m_r_solver.update_x(j, v);
     }
+
     inline void set_column_value_test(unsigned j, const impq& v) {
         set_column_value(j, v);
     }
-    unsigned get_total_iterations() const;
+    
     var_index add_named_var(unsigned ext_j, bool is_integer, const std::string&);
+
     lp_status maximize_term(unsigned j_or_term, impq &term_max);
-    inline 
-    core_solver_pretty_printer<lp::mpq, lp::impq> pp(std::ostream& out) const { return
-            core_solver_pretty_printer<lp::mpq, lp::impq>(m_mpq_lar_core_solver.m_r_solver, out); }
+
+    inline core_solver_pretty_printer<lp::mpq, lp::impq> pp(std::ostream& out) const { 
+        return core_solver_pretty_printer<lp::mpq, lp::impq>(m_mpq_lar_core_solver.m_r_solver, out); 
+    }
+
     void get_infeasibility_explanation(explanation &) const;
+
     inline void backup_x() { m_backup_x = m_mpq_lar_core_solver.m_r_x; }
+
     inline void restore_x() { m_mpq_lar_core_solver.m_r_x = m_backup_x; }
+
     template <typename T>
     void explain_implied_bound(const implied_bound & ib, lp_bound_propagator<T> & bp) {
         unsigned i = ib.m_row_or_term_index;
-        int bound_sign = ib.m_is_lower_bound? 1: -1;
-        int j_sign = (ib.m_coeff_before_j_is_pos ? 1 :-1) * bound_sign;
+        int bound_sign = (ib.m_is_lower_bound ? 1 : -1);
+        int j_sign = (ib.m_coeff_before_j_is_pos ? 1 : -1) * bound_sign;
         unsigned bound_j = ib.m_j;
-        if (tv::is_term(bound_j)) {
+        if (tv::is_term(bound_j)) 
             bound_j = m_var_register.external_to_local(bound_j);
-        }
-        for (auto const& r : A_r().m_rows[i]) {
+
+        for (auto const& r : get_row(i)) {
             unsigned j = r.var();
-            if (j == bound_j) continue;
+            if (j == bound_j) 
+                continue;
             mpq const& a = r.coeff();
-            int a_sign = is_pos(a)? 1: -1;
+            int a_sign = is_pos(a) ? 1 : -1;
             int sign = j_sign * a_sign;
             const ul_pair & ul =  m_columns_to_ul_pairs[j];
-            auto witness = sign > 0? ul.upper_bound_witness(): ul.lower_bound_witness();
+            auto witness = sign > 0 ? ul.upper_bound_witness() : ul.lower_bound_witness();
             lp_assert(is_valid(witness));
             bp.consume(a, witness);
         }
     }
+
+    void set_value_for_nbasic_column(unsigned j, const impq& new_val);
+
+    inline unsigned get_base_column_in_row(unsigned row_index) const {
+        return m_mpq_lar_core_solver.m_r_solver.get_base_column_in_row(row_index);
+    }
+
     // lp_assert(implied_bound_is_correctly_explained(ib, explanation)); }
     constraint_index mk_var_bound(var_index j, lconstraint_kind kind, const mpq & right_side);
     void activate_check_on_equal(constraint_index, var_index&);
@@ -363,9 +334,9 @@ public:
     void mark_rows_for_bound_prop(lpvar j);
     template <typename T>
     void propagate_bounds_for_touched_rows(lp_bound_propagator<T> & bp) {
-        SASSERT(use_tableau());
+        unsigned num_prop = 0;
         for (unsigned i : m_rows_with_changed_bounds) {
-            calculate_implied_bounds_for_row(i, bp);
+            num_prop += calculate_implied_bounds_for_row(i, bp);
             if (settings().get_cancel_flag())
                 return;
         }
@@ -385,8 +356,20 @@ public:
         }
         m_rows_with_changed_bounds.clear();
     }
+
+    template <typename T>
+    void check_missed_propagations(lp_bound_propagator<T> & bp) {
+        for (unsigned i = 0; i < A_r().row_count(); i++) 
+            if (!m_rows_with_changed_bounds.contains(i))                
+                if (0 < calculate_implied_bounds_for_row(i, bp)) {
+                    verbose_stream() << i << ": " << get_row(i) << "\n";
+                }
+    }
+
+    bool is_fixed_at_bound(column_index const& j);
+    bool has_fixed_at_bound();
     
-    bool is_fixed(column_index const& j) const { return column_is_fixed(j); }
+    bool is_fixed(column_index const& j) const { return column_is_fixed(j); }    
     inline column_index to_column_index(unsigned v) const { return column_index(external_to_column_index(v)); }
     bool external_is_used(unsigned) const;
     void pop(unsigned k);
@@ -416,8 +399,8 @@ public:
     void change_basic_columns_dependend_on_a_given_nb_column_report(unsigned j,
                                                                     const numeric_pair<mpq> & delta,
                                                                     const ChangeReport& after) {
-        if (use_tableau()) {
-            for (const auto & c : A_r().m_columns[j]) {
+        
+      for (const auto & c : A_r().m_columns[j]) {
                 unsigned bj = m_mpq_lar_core_solver.m_r_basis[c.var()];
                 if (tableau_with_costs()) {
                     m_basic_columns_with_changed_cost.insert(bj);
@@ -427,20 +410,8 @@ public:
                 TRACE("change_x_del",
                       tout << "changed basis column " << bj << ", it is " <<
                       ( m_mpq_lar_core_solver.m_r_solver.column_is_feasible(bj)?  "feas":"inf") << std::endl;);
-                
-                  
             }
-        } else {
-            NOT_IMPLEMENTED_YET();
-            m_column_buffer.clear();
-            m_column_buffer.resize(A_r().row_count());
-            m_mpq_lar_core_solver.m_r_solver.solve_Bd(j, m_column_buffer);
-            for (unsigned i : m_column_buffer.m_index) {
-                unsigned bj = m_mpq_lar_core_solver.m_r_basis[i];
-                m_mpq_lar_core_solver.m_r_solver.add_delta_to_x_and_track_feasibility(bj, -m_column_buffer[i] * delta); 
-            }
-        }
-    }
+   }
 
     template <typename ChangeReport>
     void set_value_for_nbasic_column_report(unsigned j,
@@ -539,7 +510,7 @@ public:
     void get_model(std::unordered_map<var_index, mpq> & variable_values) const;
     void get_rid_of_inf_eps();
     void get_model_do_not_care_about_diff_vars(std::unordered_map<var_index, mpq> & variable_values) const;
-    std::string get_variable_name(var_index vi) const;
+    std::string get_variable_name(var_index vi) const override;
     void set_variable_name(var_index vi, std::string);
     inline unsigned number_of_vars() const { return m_var_register.size(); }
     inline bool is_base(unsigned j) const { return m_mpq_lar_core_solver.m_r_heading[j] >= 0; }
@@ -547,8 +518,6 @@ public:
         return m_mpq_lar_core_solver.lower_bound(j);
     }
 
-    void pivot_column_tableau(unsigned j, unsigned row_index);
-    
     inline const impq & column_upper_bound(unsigned j) const {
         return m_mpq_lar_core_solver.upper_bound(j);
     }
@@ -627,6 +596,7 @@ public:
     }
     void round_to_integer_solution();
     inline const row_strip<mpq> &  get_row(unsigned i) const { return A_r().m_rows[i]; }
+    inline const row_strip<mpq> &  basic2row(unsigned i) const { return A_r().m_rows[row_of_basic_column(i)]; }
     inline const column_strip &  get_column(unsigned i) const { return A_r().m_columns[i]; }
     bool row_is_correct(unsigned i) const;
     bool ax_is_correct() const;
@@ -646,7 +616,7 @@ public:
     lar_solver();
     void set_track_pivoted_rows(bool v);
     bool get_track_pivoted_rows() const;    
-    virtual ~lar_solver();
+    ~lar_solver() override;
     const vector<impq>& r_x() const { return m_mpq_lar_core_solver.m_r_x; }
     bool column_is_int(unsigned j) const;
     inline bool column_value_is_int(unsigned j) const { return m_mpq_lar_core_solver.m_r_x[j].is_int(); }
