@@ -9,7 +9,7 @@
    
     DDFW Local search module for clauses
 
-  Author:
+  Author:        
 
     Nikolaj Bjorner, Marijn Heule 2019-4-23
      
@@ -33,102 +33,164 @@
 namespace sat {
 
     ddfw::~ddfw() {
-        for (auto& ci : m_clauses) {
-            m_alloc.del_clause(ci.m_clause);
-        }
+        for (auto& ci : m_clauses) 
+            m_alloc.del_clause(ci.m_clause);        
     }
 
-
     lbool ddfw::check(unsigned sz, literal const* assumptions, parallel* p) {
         init(sz, assumptions);
         flet<parallel*> _p(m_par, p);
+        if (m_plugin)
+            check_with_plugin();
+        else
+            check_without_plugin();
+        remove_assumptions();
+        log();
+        return m_min_sz == 0 ? l_true : l_undef;
+    }
+
+    void ddfw::check_without_plugin() {
         while (m_limit.inc() && m_min_sz > 0) {
             if (should_reinit_weights()) do_reinit_weights();
-            else if (do_flip()) ;
+            else if (do_flip<false>());
             else if (should_restart()) do_restart();
             else if (should_parallel_sync()) do_parallel_sync();
-            else shift_weights();                       
+            else shift_weights();
         }
-        return m_min_sz == 0 ? l_true : l_undef;
+    }
+
+    void ddfw::check_with_plugin() {
+        m_plugin->init_search();
+        m_steps_since_progress = 0;
+        unsigned steps = 0;
+        while (m_min_sz > 0 && m_steps_since_progress++ <= 1500000) {
+            if (should_reinit_weights()) do_reinit_weights();
+            else if (steps % 5000 == 0) shift_weights(), m_plugin->on_rescale();
+            else if (should_restart()) do_restart(), m_plugin->on_restart();
+            else if (do_flip<true>());
+            else if (do_literal_flip<true>());
+            else if (should_parallel_sync()) do_parallel_sync();
+            else shift_weights(), m_plugin->on_rescale();
+            ++steps;
+        }
+        m_plugin->finish_search();
     }
 
     void ddfw::log() {
         double sec = m_stopwatch.get_current_seconds();        
         double kflips_per_sec = (m_flips - m_last_flips) / (1000.0 * sec);
         if (m_last_flips == 0) {
-            IF_VERBOSE(0, verbose_stream() << "(sat.ddfw :unsat :models :kflips/sec  :flips  :restarts  :reinits  :unsat_vars  :shifts";
+            IF_VERBOSE(1, verbose_stream() << "(sat.ddfw :unsat :models :kflips/sec  :flips  :restarts  :reinits  :unsat_vars  :shifts";
                        if (m_par) verbose_stream() << "  :par";
                        verbose_stream() << ")\n");
         }
-        IF_VERBOSE(0, verbose_stream() << "(sat.ddfw " 
+        IF_VERBOSE(1, verbose_stream() << "(sat.ddfw " 
                    << std::setw(07) << m_min_sz 
                    << std::setw(07) << m_models.size()
                    << std::setw(10) << kflips_per_sec
                    << std::setw(10) << m_flips 
                    << std::setw(10) << m_restart_count
-                   << std::setw(10) << m_reinit_count
-                   << std::setw(10) << m_unsat_vars.size()
-                   << std::setw(10) << m_shifts;
+                   << std::setw(11) << m_reinit_count
+                   << std::setw(13) << m_unsat_vars.size()
+                   << std::setw(9) << m_shifts;
                    if (m_par) verbose_stream() << std::setw(10) << m_parsync_count;
                    verbose_stream() << ")\n");
         m_stopwatch.start();
         m_last_flips = m_flips;
     }
 
+    template<bool uses_plugin>
     bool ddfw::do_flip() {
-        bool_var v = pick_var();
-        if (reward(v) > 0 || (reward(v) == 0 && m_rand(100) <= m_config.m_use_reward_zero_pct)) {
-            flip(v);
-            if (m_unsat.size() <= m_min_sz) save_best_values();
+        double reward = 0;
+        bool_var v = pick_var<uses_plugin>(reward);
+        return apply_flip<uses_plugin>(v, reward);
+    }
+
+    template<bool uses_plugin>
+    bool ddfw::apply_flip(bool_var v, double reward) {
+        if (v == null_bool_var) 
+            return false;
+        
+        if (reward > 0 || (reward == 0 && m_rand(100) <= m_config.m_use_reward_zero_pct)) {
+            if (uses_plugin && is_external(v))
+                m_plugin->flip(v);
+            else
+                flip(v);
+            if (m_unsat.size() <= m_min_sz) 
+                save_best_values();
             return true;
         }
         return false;
     }
 
-    bool_var ddfw::pick_var() {
+    template<bool uses_plugin>
+    bool_var ddfw::pick_var(double& r) {
         double sum_pos = 0;
         unsigned n = 1;
         bool_var v0 = null_bool_var;
         for (bool_var v : m_unsat_vars) {
-            int r = reward(v);
-            if (r > 0) {                
-                sum_pos += score(r);
-            }
-            else if (r == 0 && sum_pos == 0 && (m_rand() % (n++)) == 0) {
-                v0 = v;
-            }
+            r = uses_plugin ? plugin_reward(v) : reward(v);
+            if (r > 0.0)    
+                sum_pos += score(r);            
+            else if (r == 0.0 && sum_pos == 0 && (m_rand() % (n++)) == 0) 
+                v0 = v;            
         }
         if (sum_pos > 0) {
             double lim_pos = ((double) m_rand() / (1.0 + m_rand.max_value())) * sum_pos;                
             for (bool_var v : m_unsat_vars) {
-                int r = reward(v);
+                r = uses_plugin && is_external(v) ? m_vars[v].m_last_reward : reward(v);
                 if (r > 0) {
                     lim_pos -= score(r);
-                    if (lim_pos <= 0) {
-                        if (m_par) update_reward_avg(v);
-                        return v;
-                    }
+                    if (lim_pos <= 0) 
+                        return v;                    
                 }
             }
         }
-        if (v0 != null_bool_var) {
+        r = 0;
+        if (v0 != null_bool_var) 
             return v0;
-        }
+        if (m_unsat_vars.empty())
+            return null_bool_var;
         return m_unsat_vars.elem_at(m_rand(m_unsat_vars.size()));
     }
 
-    /**
-     * TBD: map reward value to a score, possibly through an exponential function, such as
-     * exp(-tau/r), where tau > 0
-     */
-    double ddfw::mk_score(unsigned r) {
-        return r;
+    template<bool uses_plugin>
+    bool ddfw::do_literal_flip() {
+        double reward = 1;
+        return apply_flip<uses_plugin>(pick_literal_var<uses_plugin>(), reward);
     }
 
+    /*
+    * Pick a random false literal from a satisfied clause such that
+    * the literal has zero break count and positive reward.
+    */
+    template<bool uses_plugin>
+    bool_var ddfw::pick_literal_var() {
+#if false
+        unsigned sz = m_clauses.size();
+        unsigned start = rand();
+        for (unsigned i = 0; i < 100; ++i) {
+            unsigned cl = (i + start) % sz;
+            if (m_unsat.contains(cl))
+                continue;
+            for (auto lit : *m_clauses[cl].m_clause) {
+                if (is_true(lit))
+                    continue;
+                double r = uses_plugin ? plugin_reward(lit.var()) : reward(lit.var());
+                if (r < 0)
+                    continue;
+                //verbose_stream() << "false " << r << " " << lit << "\n";
+                return lit.var();
+            }
+        }
+#endif
+        return null_bool_var;
+    }
 
     void ddfw::add(unsigned n, literal const* c) {        
         clause* cls = m_alloc.mk_clause(n, c, false);
         unsigned idx = m_clauses.size();
+
         m_clauses.push_back(clause_info(cls, m_config.m_init_clause_weight));
         for (literal lit : *cls) {
             m_use_list.reserve(2*(lit.var()+1));
@@ -137,10 +199,22 @@ namespace sat {
         }
     }
 
+    /**
+     * Remove the last clause that was added
+     */
+    void ddfw::del() {
+        auto& info = m_clauses.back();
+        for (literal lit : *info.m_clause)
+            m_use_list[lit.index()].pop_back();
+        m_alloc.del_clause(info.m_clause);
+        m_clauses.pop_back();
+        if (m_unsat.contains(m_clauses.size()))
+            m_unsat.remove(m_clauses.size());
+    }
+
     void ddfw::add(solver const& s) {
-        for (auto& ci : m_clauses) {
+        for (auto& ci : m_clauses) 
             m_alloc.del_clause(ci.m_clause);
-        }
         m_clauses.reset(); 
         m_use_list.reset();
         m_num_non_binary_clauses = 0;
@@ -170,9 +244,16 @@ namespace sat {
     }
 
     void ddfw::add_assumptions() {
-        for (unsigned i = 0; i < m_assumptions.size(); ++i) {
-            add(1, m_assumptions.data() + i);
-        }
+        for (unsigned i = 0; i < m_assumptions.size(); ++i) 
+            add(1, m_assumptions.data() + i);        
+    }
+
+    void ddfw::remove_assumptions() {
+        if (m_assumptions.empty())
+            return;
+        for (unsigned i = 0; i < m_assumptions.size(); ++i) 
+            del();        
+        init(0, nullptr);
     }
 
     void ddfw::init(unsigned sz, literal const* assumptions) {
@@ -202,15 +283,13 @@ namespace sat {
         m_stopwatch.start();
     }
 
-    void ddfw::reinit(solver& s) {
+    void ddfw::reinit(solver& s, bool_vector const& phase) {
         add(s);
         add_assumptions();
-        if (s.m_best_phase_size > 0) {
-            for (unsigned v = 0; v < num_vars(); ++v) {                
-                value(v) = s.m_best_phase[v];
-                reward(v) = 0;
-                make_count(v) = 0;                
-            }
+        for (unsigned v = 0; v < phase.size(); ++v) {
+            value(v) = phase[v];
+            reward(v) = 0;
+            make_count(v) = 0;
         }
         init_clause_data();
         flatten_use_list();
@@ -226,7 +305,6 @@ namespace sat {
         m_use_list_index.push_back(m_flat_use_list.size());
     }
 
-
     void ddfw::flip(bool_var v) {
         ++m_flips;
         literal lit = literal(v, !value(v));
@@ -235,7 +313,7 @@ namespace sat {
         for (unsigned cls_idx : use_list(*this, lit)) {
             clause_info& ci = m_clauses[cls_idx];
             ci.del(lit);
-            unsigned w = ci.m_weight;
+            double w = ci.m_weight;
             // cls becomes false: flip any variable in clause to receive reward w
             switch (ci.m_num_trues) {
             case 0: {
@@ -257,7 +335,7 @@ namespace sat {
         }
         for (unsigned cls_idx : use_list(*this, nlit)) {
             clause_info& ci = m_clauses[cls_idx];             
-            unsigned w = ci.m_weight;
+            double w = ci.m_weight;
             // the clause used to have a single true (pivot) literal, now it has two.
             // Then the previous pivot is no longer penalized for flipping.
             switch (ci.m_num_trues) {
@@ -280,6 +358,7 @@ namespace sat {
             ci.add(nlit);
         }
         value(v) = !value(v);
+        update_reward_avg(v);
     }
 
     bool ddfw::should_reinit_weights() {
@@ -290,19 +369,15 @@ namespace sat {
         log();
 
         if (m_reinit_count % 2 == 0) { 
-            for (auto& ci : m_clauses) {
-                ci.m_weight += 1;                
-            }
+            for (auto& ci : m_clauses) 
+                ci.m_weight += 1;                            
         }
         else {
-            for (auto& ci : m_clauses) {
-                if (ci.is_true()) {
-                    ci.m_weight = m_config.m_init_clause_weight;
-                }
-                else {
-                    ci.m_weight = m_config.m_init_clause_weight + 1;
-                }                
-            }
+            for (auto& ci : m_clauses) 
+                if (ci.is_true()) 
+                    ci.m_weight = m_config.m_init_clause_weight;                
+                else 
+                    ci.m_weight = m_config.m_init_clause_weight + 1;                                       
         }
         init_clause_data();   
         ++m_reinit_count;
@@ -322,11 +397,9 @@ namespace sat {
             clause const& c = get_clause(i);
             ci.m_trues = 0;
             ci.m_num_trues = 0;
-            for (literal lit : c) {
-                if (is_true(lit)) {
-                    ci.add(lit);
-                }
-            }
+            for (literal lit : c) 
+                if (is_true(lit)) 
+                    ci.add(lit);                            
             switch (ci.m_num_trues) {
             case 0:
                 for (literal lit : c) {
@@ -347,7 +420,7 @@ namespace sat {
     bool ddfw::should_restart() {
         return m_flips >= m_restart_next;
     }
-
+    
     void ddfw::do_restart() {        
         reinit_values();
         init_clause_data();
@@ -365,12 +438,10 @@ namespace sat {
     void ddfw::reinit_values() {
         for (unsigned i = 0; i < num_vars(); ++i) {
             int b = bias(i);
-            if (0 == (m_rand() % (1 + abs(b)))) {
-                value(i) = (m_rand() % 2) == 0;
-            }
-            else {
-                value(i) = bias(i) > 0;
-            }
+            if (0 == (m_rand() % (1 + abs(b)))) 
+                value(i) = (m_rand() % 2) == 0;            
+            else 
+                value(i) = bias(i) > 0;            
         }        
     }
 
@@ -378,37 +449,36 @@ namespace sat {
         return m_par != nullptr && m_flips >= m_parsync_next;
     }
 
+    void ddfw::save_priorities() {
+        m_probs.reset();
+        for (unsigned v = 0; v < num_vars(); ++v) 
+            m_probs.push_back(-m_vars[v].m_reward_avg);         
+    }
+
     void ddfw::do_parallel_sync() {
-        if (m_par->from_solver(*this)) {
-            // Sum exp(xi) / exp(a) = Sum exp(xi - a)
-            double max_avg = 0;
-            for (unsigned v = 0; v < num_vars(); ++v) {
-                max_avg = std::max(max_avg, (double)m_vars[v].m_reward_avg);
-            }
-            double sum = 0;
-            for (unsigned v = 0; v < num_vars(); ++v) {
-                sum += exp(m_config.m_itau * (m_vars[v].m_reward_avg - max_avg));
-            }
-            if (sum == 0) {
-                sum = 0.01;
-            }
-            m_probs.reset();
-            for (unsigned v = 0; v < num_vars(); ++v) {
-                m_probs.push_back(exp(m_config.m_itau * (m_vars[v].m_reward_avg - max_avg)) / sum);
-            }
+        if (m_par->from_solver(*this)) 
             m_par->to_solver(*this);
-        }
+        
         ++m_parsync_count;
         m_parsync_next *= 3;
         m_parsync_next /= 2;
     }
 
+    void ddfw::save_model() {
+        m_model.reserve(num_vars());
+        for (unsigned i = 0; i < num_vars(); ++i) 
+            m_model[i] = to_lbool(value(i));
+        save_priorities();
+        if (m_plugin)
+            m_plugin->on_save_model();
+    }
+
+
     void ddfw::save_best_values() {
-        if (m_unsat.empty()) {
-            m_model.reserve(num_vars());
-            for (unsigned i = 0; i < num_vars(); ++i) {
-                m_model[i] = to_lbool(value(i));
-            }
+        if (m_unsat.size() < m_min_sz) {
+            m_steps_since_progress = 0;
+            if (m_unsat.size() < 50 || m_min_sz * 10 > m_unsat.size() * 11)
+                save_model();
         }
         if (m_unsat.size() < m_min_sz) {
             m_models.reset();
@@ -420,15 +490,20 @@ namespace sat {
                 }
             }
         }
+
         unsigned h = value_hash();
+        unsigned occs = 0;
+        bool contains = m_models.find(h, occs);
         if (!m_models.contains(h)) {
-            for (unsigned v = 0; v < num_vars(); ++v) {
+            for (unsigned v = 0; v < num_vars(); ++v)
                 bias(v) += value(v) ? 1 : -1;
-            }
-            m_models.insert(h);
-            if (m_models.size() > m_config.m_max_num_models) {                
-                m_models.erase(*m_models.begin());
-            }
+            if (m_models.size() > m_config.m_max_num_models)
+                m_models.erase(m_models.begin()->m_key);
+        }
+        m_models.insert(h, occs + 1);
+        if (occs > 100) {
+            m_restart_next = m_flips;            
+            m_models.erase(h);
         }
         m_min_sz = m_unsat.size();
     }
@@ -450,10 +525,9 @@ namespace sat {
        3. select multiple clauses instead of just one per clause in unsat.
      */
 
-    bool ddfw::select_clause(unsigned max_weight, unsigned max_trues, clause_info const& cn, unsigned& n) {
-        if (cn.m_num_trues == 0 || cn.m_weight < max_weight) {
+    bool ddfw::select_clause(double max_weight, clause_info const& cn, unsigned& n) {
+        if (cn.m_num_trues == 0 || cn.m_weight + 1e-5 < max_weight) 
             return false;
-        }
         if (cn.m_weight > max_weight) {
             n = 2;
             return true;
@@ -462,51 +536,71 @@ namespace sat {
     }
 
     unsigned ddfw::select_max_same_sign(unsigned cf_idx) {
-        clause const& c = get_clause(cf_idx);
-        unsigned max_weight = 2;
-        unsigned max_trues = 0;
+        auto& ci = m_clauses[cf_idx];
         unsigned cl = UINT_MAX; // clause pointer to same sign, max weight satisfied clause.
+        clause const& c = *ci.m_clause;
+        double max_weight = m_init_weight;
         unsigned n = 1;
         for (literal lit : c) {
             for (unsigned cn_idx : use_list(*this, lit)) {
                 auto& cn = m_clauses[cn_idx];
-                if (select_clause(max_weight, max_trues, cn, n)) {
+                if (select_clause(max_weight, cn, n)) {
                     cl = cn_idx;
                     max_weight = cn.m_weight;
-                    max_trues = cn.m_num_trues;
                 }
             }
         }
         return cl;
     }
 
+    void ddfw::transfer_weight(unsigned from, unsigned to, double w) {
+        auto& cf = m_clauses[to];
+        auto& cn = m_clauses[from];
+        if (cn.m_weight < w) 
+            return;
+        cf.m_weight += w;
+        cn.m_weight -= w;
+        
+        for (literal lit : get_clause(to)) 
+            inc_reward(lit, w);
+        if (cn.m_num_trues == 1) 
+            inc_reward(to_literal(cn.m_trues), w);
+    }
+
+    unsigned ddfw::select_random_true_clause() {
+        unsigned num_clauses = m_clauses.size();
+        unsigned rounds = 100 * num_clauses;
+        for (unsigned i = 0; i < rounds; ++i) {
+            unsigned idx = (m_rand() * m_rand()) % num_clauses;
+            auto & cn = m_clauses[idx];
+            if (cn.is_true() && cn.m_weight >= m_init_weight) 
+                return idx;
+        }
+        return UINT_MAX;
+    }
+
+    // 1% chance to disregard neighbor
+    inline bool ddfw::disregard_neighbor() {
+        return false; // rand() % 1000 == 0;
+    }
+
+    double ddfw::calculate_transfer_weight(double w) {
+        return (w > m_init_weight) ? m_init_weight : 1; 
+    }
+
     void ddfw::shift_weights() {
         ++m_shifts;
-        for (unsigned cf_idx : m_unsat) {
-            auto& cf = m_clauses[cf_idx];
-            SASSERT(!cf.is_true());
-            unsigned cn_idx = select_max_same_sign(cf_idx);
-            while (cn_idx == UINT_MAX) {
-                unsigned idx = (m_rand() * m_rand()) % m_clauses.size();
-                auto & cn = m_clauses[idx];
-                if (cn.is_true() && cn.m_weight >= 2) {
-                    cn_idx = idx;
-                }
-            }
-            auto & cn = m_clauses[cn_idx];
+        for (unsigned to_idx : m_unsat) {
+            SASSERT(!m_clauses[to_idx].is_true());
+            unsigned from_idx = select_max_same_sign(to_idx);
+            if (from_idx == UINT_MAX || disregard_neighbor())
+                from_idx = select_random_true_clause();
+            if (from_idx == UINT_MAX)
+                continue;
+            auto & cn = m_clauses[from_idx];
             SASSERT(cn.is_true());
-            unsigned wn = cn.m_weight;
-            SASSERT(wn >= 2);
-            unsigned inc = (wn > 2) ? 2 : 1; 
-            SASSERT(wn - inc >= 1);            
-            cf.m_weight += inc;
-            cn.m_weight -= inc;
-            for (literal lit : get_clause(cf_idx)) {
-                inc_reward(lit, inc);
-            }
-            if (cn.m_num_trues == 1) {
-                inc_reward(to_literal(cn.m_trues), inc);
-            }
+            double w = calculate_transfer_weight(cn.m_weight);
+            transfer_weight(from_idx, to_idx, w);
         }
         // DEBUG_CODE(invariant(););
     }
@@ -543,7 +637,7 @@ namespace sat {
             VERIFY(found);
         }
         for (unsigned v = 0; v < num_vars(); ++v) {
-            int v_reward = 0;
+            double v_reward = 0;
             literal lit(v, !value(v));
             for (unsigned j : m_use_list[lit.index()]) {
                 clause_info const& ci = m_clauses[j];
@@ -559,7 +653,7 @@ namespace sat {
                 }                
             }
             IF_VERBOSE(0, if (v_reward != reward(v)) verbose_stream() << v << " " << v_reward << " " << reward(v) << "\n");
-            SASSERT(reward(v) == v_reward);
+            // SASSERT(reward(v) == v_reward);
         }
         DEBUG_CODE(
             for (auto const& ci : m_clauses) {
