@@ -69,9 +69,12 @@ namespace q {
             [&](euf::enode* n) {
             m_mam->add_node(n, false);
         };
-        ctx.get_egraph().set_on_merge(_on_merge);
-        if (!ctx.relevancy_enabled())
-            ctx.get_egraph().set_on_make(_on_make);
+        
+        if (ctx.get_config().m_ematching) {
+            ctx.get_egraph().set_on_merge(_on_merge);
+            if (!ctx.relevancy_enabled())
+                ctx.get_egraph().set_on_make(_on_make);
+        }
         m_mam = mam::mk(ctx, *this);
     }
 
@@ -104,7 +107,7 @@ namespace q {
     * is created to ensure the justification trail is well-founded
     * during conflict resolution.
     */
-    sat::ext_justification_idx ematch::mk_justification(unsigned idx, clause& c, euf::enode* const* b) {
+    sat::ext_justification_idx ematch::mk_justification(unsigned idx, unsigned generation, clause& c, euf::enode* const* b) {
         void* mem = ctx.get_region().allocate(justification::get_obj_size());
         sat::constraint_base::initialize(mem, &m_qs);
         bool sign = false;
@@ -113,21 +116,23 @@ namespace q {
         if (idx != UINT_MAX)
             lit = c[idx];
         m_explain.reset();
+        m_explain_cc.reset();
         ctx.get_egraph().begin_explain();
         ctx.reset_explain();
+        euf::cc_justification* cc = ctx.use_drat() ? &m_explain_cc : nullptr;
         for (auto const& [a, b] : m_evidence) {
             SASSERT(a->get_root() == b->get_root() || ctx.get_egraph().are_diseq(a, b));
             if (a->get_root() == b->get_root())
-                ctx.get_egraph().explain_eq<size_t>(m_explain, a, b);
+                ctx.get_egraph().explain_eq<size_t>(m_explain, cc, a, b);
             else
-                ctx.add_diseq_antecedent(m_explain, a, b);
+                ctx.add_diseq_antecedent(m_explain, cc, a, b);
         }
         ctx.get_egraph().end_explain();
 
         size_t** ev = static_cast<size_t**>(ctx.get_region().allocate(sizeof(size_t*) * m_explain.size()));
         for (unsigned i = m_explain.size(); i-- > 0; )
             ev[i] = m_explain[i];
-        auto* constraint = new (sat::constraint_base::ptr2mem(mem)) justification(lit, c, b, m_explain.size(), ev);
+        auto* constraint = new (sat::constraint_base::ptr2mem(mem)) justification(lit, c, b, generation, m_explain.size(), ev);
         return constraint->to_index();
     }
 
@@ -341,7 +346,6 @@ namespace q {
             return false;
         if (ctx.s().inconsistent())
             return true;
-        TRACE("q", c.display(ctx, tout) << "\n";);
         unsigned idx = UINT_MAX;
         m_evidence.reset();
         lbool ev = m_eval(binding, c, idx, m_evidence);
@@ -362,7 +366,7 @@ namespace q {
         if (!is_owned) 
             binding = copy_nodes(c, binding); 
 
-        auto j_idx = mk_justification(idx, c, binding);     
+        auto j_idx = mk_justification(idx, max_generation, c, binding);     
 
         if (is_owned)
             propagate(ev == l_false, idx, j_idx);
@@ -373,16 +377,21 @@ namespace q {
     }
 
     void ematch::propagate(bool is_conflict, unsigned idx, sat::ext_justification_idx j_idx) {
-        if (is_conflict) {
+        if (is_conflict) 
             ++m_stats.m_num_conflicts;
-            ctx.set_conflict(j_idx);
-        }
-        else {
+        else
             ++m_stats.m_num_propagations;
-            auto& j = justification::from_index(j_idx);
-            auto lit = instantiate(j.m_clause, j.m_binding, j.m_clause[idx]);
-            ctx.propagate(lit, j_idx);
-        }
+
+        auto& j = justification::from_index(j_idx);
+        sat::literal_vector lits;
+        lits.push_back(~j.m_clause.m_literal);
+        for (unsigned i = 0; i < j.m_clause.size(); ++i) 
+            lits.push_back(instantiate(j.m_clause, j.m_generation, j.m_binding, j.m_clause[i])); 
+        m_qs.log_instantiation(lits, &j);
+        euf::th_proof_hint* ph = nullptr;
+        if (ctx.use_drat()) 
+            ph = q_proof_hint::mk(ctx, j.m_generation, lits, j.m_clause.num_decls(), j.m_binding);
+        m_qs.add_clause(lits, ph);               
     }
 
     bool ematch::flush_prop_queue() {
@@ -408,14 +417,16 @@ namespace q {
 
     void ematch::add_instantiation(clause& c, binding& b, sat::literal lit) {
         m_evidence.reset();
-        ctx.propagate(lit, mk_justification(UINT_MAX, c, b.nodes()));
+        ctx.propagate(lit, mk_justification(UINT_MAX, b.m_max_generation, c, b.nodes()));
+        m_qs.log_instantiation(~c.m_literal, lit);
     }
 
-    sat::literal ematch::instantiate(clause& c, euf::enode* const* binding, lit const& l) {
+    sat::literal ematch::instantiate(clause& c, unsigned generation, euf::enode* const* binding, lit const& l) {
         expr_ref_vector _binding(m);
         for (unsigned i = 0; i < c.num_decls(); ++i)
             _binding.push_back(binding[i]->get_expr());
         var_subst subst(m);
+        euf::solver::scoped_generation sg(ctx, generation + 1);
         auto sub = [&](expr* e) {
             expr_ref r = subst(e, _binding);
             //ctx.rewrite(r);
@@ -611,39 +622,49 @@ namespace q {
         return ctx.get_config().m_ematching && propagate(false);
     }
 
+    void ematch::propagate(clause& c, bool flush, bool& propagated) {
+        ptr_buffer<binding> to_remove;
+        binding* b = c.m_bindings;
+        if (!b)
+            return;
+
+        do {                
+            if (propagate(true, b->m_nodes, b->m_max_generation, c, propagated)) 
+                to_remove.push_back(b);
+            else if (flush) {
+                instantiate(*b);
+                to_remove.push_back(b);
+                propagated = true;
+            }
+            b = b->next();
+        } 
+        while (b != c.m_bindings);
+        
+        for (auto* b : to_remove) {
+            SASSERT(binding::contains(c.m_bindings, b));
+            binding::remove_from(c.m_bindings, b);
+            binding::detach(b);
+            ctx.push(insert_binding(ctx, c, b));
+        }
+    }
+
+
     bool ematch::propagate(bool flush) {
         m_mam->propagate();
         bool propagated = flush_prop_queue();
-        if (m_qhead >= m_clause_queue.size())
-            return m_inst_queue.propagate() || propagated;
-        ctx.push(value_trail<unsigned>(m_qhead));
-        ptr_buffer<binding> to_remove;
-        for (; m_qhead < m_clause_queue.size() && m.inc(); ++m_qhead) {
-            unsigned idx = m_clause_queue[m_qhead];
-            clause& c = *m_clauses[idx];
-            binding* b = c.m_bindings;
-            if (!b)
-                continue;
-
-            do {                
-                if (propagate(true, b->m_nodes, b->m_max_generation, c, propagated)) 
-                    to_remove.push_back(b);
-                else if (flush) {
-                    instantiate(*b);
-                    to_remove.push_back(b);
-                    propagated = true;
-                }
-                b = b->next();
-            } 
-            while (b != c.m_bindings);
-
-            for (auto* b : to_remove) {
-                SASSERT(binding::contains(c.m_bindings, b));
-                binding::remove_from(c.m_bindings, b);
-                binding::detach(b);
-                ctx.push(insert_binding(ctx, c, b));
+        if (flush) {
+            for (auto* c : m_clauses)
+                propagate(*c, flush, propagated);
+        }
+        else {
+            if (m_qhead >= m_clause_queue.size())
+                return m_inst_queue.propagate() || propagated;
+            ctx.push(value_trail<unsigned>(m_qhead));
+            for (; m_qhead < m_clause_queue.size() && m.inc(); ++m_qhead) {
+                unsigned idx = m_clause_queue[m_qhead];
+                clause& c = *m_clauses[idx];
+                propagate(c, flush, propagated);
             }
-            to_remove.reset();
         }
         m_clause_in_queue.reset();
         m_node_in_queue.reset();
@@ -662,15 +683,18 @@ namespace q {
         if (propagate(false))
             return true;        
         for (unsigned i = 0; i < m_clauses.size(); ++i)
-            if (m_clauses[i]->m_bindings)
+            if (m_clauses[i]->m_bindings) 
                 insert_clause_in_queue(i);
         if (propagate(true))
             return true;
         if (m_inst_queue.lazy_propagate())
             return true;
         for (unsigned i = 0; i < m_clauses.size(); ++i)
-            if (m_clauses[i]->m_bindings)
+            if (m_clauses[i]->m_bindings) {
                 IF_VERBOSE(0, verbose_stream() << "missed propagation " << i << "\n");
+                TRACE("q", display(tout << "missed propagation\n"));
+                break;
+            }
         
         TRACE("q", tout << "no more propagation\n";);
         return false;
