@@ -54,6 +54,20 @@ namespace smt {
         return r;
     }
 
+    app * theory_array_base::mk_select_reduce(unsigned num_args, expr * * args) {
+        array_util util(m);
+        while (util.is_store(args[0])) {
+            bool are_distinct = false;
+            for (unsigned i = 1; i < num_args && !are_distinct; ++i) 
+                are_distinct |= m.are_distinct(args[i], to_app(args[0])->get_arg(i));
+            if (!are_distinct)
+                break;
+            args[0] = to_app(to_app(args[0])->get_arg(0));
+        }
+        return mk_select(num_args, args);
+    }
+
+
     app * theory_array_base::mk_store(unsigned num_args, expr * const * args) {
         return m.mk_app(get_family_id(), OP_STORE, 0, nullptr, num_args, args);
     }
@@ -101,7 +115,7 @@ namespace smt {
         SASSERT(num_args >= 3);
         sel_args.push_back(n);
         for (unsigned i = 1; i < num_args - 1; ++i) {
-            sel_args.push_back(to_app(n->get_arg(i)));
+            sel_args.push_back(n->get_arg(i));
         }
         expr_ref sel(m);
         sel = mk_select(sel_args.size(), sel_args.data());
@@ -179,6 +193,12 @@ namespace smt {
                 conseq_expr = ctx.bool_var2expr(conseq.var());
             }
 
+            if (m.are_distinct(idx1->get_expr(), idx2->get_expr())) {
+                ctx.mark_as_relevant(conseq);
+                assert_axiom(conseq);
+                continue;
+            }
+
             literal ante = mk_eq(idx1->get_expr(), idx2->get_expr(), true);
             ctx.mark_as_relevant(ante);
             // ctx.force_phase(ante);
@@ -473,39 +493,23 @@ namespace smt {
         return false;
     }
 
-#if 0
-    void theory_array_base::collect_shared_vars(sbuffer<theory_var> & result) {
-        TRACE("array_shared", tout << "collecting shared vars...\n";);
-        ptr_buffer<enode> to_unmark;
-        unsigned num_vars = get_num_vars();
-        for (unsigned i = 0; i < num_vars; i++) {
-            enode * n = get_enode(i);
-            if (ctx.is_relevant(n) && ctx.is_shared(n)) {
-                enode * r = n->get_root();
-                if (!r->is_marked() && is_array_sort(r)) {
-                    TRACE("array_shared", tout << "new shared var: #" << r->get_expr_id() << "\n";);
-                    r->set_mark();
-                    to_unmark.push_back(r);
-                    theory_var r_th_var = r->get_var(get_id());
-                    SASSERT(r_th_var != null_theory_var);
-                    result.push_back(r_th_var);
-                }
-            }
-        }
-        unmark_enodes(to_unmark.size(), to_unmark.c_ptr());
+    bool theory_array_base::is_beta_redex(enode* p, enode* n) const {
+        if (is_select(p))
+            return p->get_arg(0)->get_root() == n->get_root();
+        if (is_map(p))
+            return true;
+        if (is_store(p))
+            return true;
+        return false;
     }
-#else
+
 
     bool theory_array_base::is_select_arg(enode* r) {
-        for (enode* n : r->get_parents()) {
-            if (is_select(n)) {
-                for (unsigned i = 1; i < n->get_num_args(); ++i) {
-                    if (r == n->get_arg(i)->get_root()) {
+        for (enode* n : r->get_parents()) 
+            if (is_select(n)) 
+                for (unsigned i = 1; i < n->get_num_args(); ++i) 
+                    if (r == n->get_arg(i)->get_root()) 
                         return true;
-                    }
-                }
-            }
-        }
         return false;
     }
 
@@ -525,7 +529,7 @@ namespace smt {
             // issue #3532, #3529
             // 
             if (ctx.is_shared(r) || is_select_arg(r)) {
-                TRACE("array", tout << "new shared var: #" << r->get_owner_id() << "\n";);
+                TRACE("array", tout << "new shared var: #" << r->get_owner_id() << " " << is_select_arg(r) << "\n";);
                 theory_var r_th_var = r->get_th_var(get_id());
                 SASSERT(r_th_var != null_theory_var);
                 result.push_back(r_th_var);
@@ -536,7 +540,6 @@ namespace smt {
         TRACE("array", tout << "collecting shared vars...\n" << unsigned_vector(result.size(), (unsigned*)result.data())  << "\n";);
         unmark_enodes(to_unmark.size(), to_unmark.data());
     }
-#endif
 
     /**
        \brief Create interface variables for shared array variables.
@@ -909,8 +912,6 @@ namespace smt {
             m_unspecified_else(true) {
         }
 
-        ~array_value_proc() override {}
-     
         void add_entry(unsigned num_args, enode * const * args, enode * value) {
             SASSERT(num_args > 0);
             SASSERT(m_dim == 0 || m_dim == num_args);
