@@ -252,6 +252,8 @@ namespace smt {
         else if (m.is_lambda_def(n->get_decl())) {
             instantiate_default_lambda_def_axiom(n);
             d->m_lambdas.push_back(n);
+            m_lambdas.push_back(n);
+            ctx.push_trail(push_back_vector(m_lambdas));
         }
         return r;
     }
@@ -330,7 +332,8 @@ namespace smt {
             // Even if there was, as-array on interpreted 
             // functions will be incomplete.
             // The instantiation operations are still sound to include.
-            found_unsupported_op(n);
+            m_as_array.push_back(node);
+            ctx.push_trail(push_back_vector(m_as_array));
             instantiate_default_as_array_axiom(node);
         }
         else if (is_array_ext(n)) {
@@ -594,13 +597,24 @@ namespace smt {
         if (!ctx.add_fingerprint(this, m_default_lambda_fingerprint, 1, &arr))
             return false;
         m_stats.m_num_default_lambda_axiom++;
-        expr* def = mk_default(arr->get_expr());
+        expr* e = arr->get_expr();
+        expr* def = mk_default(e);
         quantifier* lam = m.is_lambda_def(arr->get_decl());
-        expr_ref_vector args(m);
-        args.push_back(lam);
+        TRACE("array", tout << mk_pp(lam, m) << "\n" << mk_pp(e, m) << "\n");
+        expr_ref_vector args(m);       
+        var_subst subst(m, false);
+        args.push_back(subst(lam, to_app(e)->get_num_args(), to_app(e)->get_args()));
         for (unsigned i = 0; i < lam->get_num_decls(); ++i) 
             args.push_back(mk_epsilon(lam->get_decl_sort(i)).first);
         expr_ref val(mk_select(args), m);
+        ctx.get_rewriter()(val);
+        if (has_quantifiers(val)) {
+            expr_ref fn(m.mk_fresh_const("lambda-body", val->get_sort()), m);
+            expr_ref eq(m.mk_eq(fn, val), m);
+            ctx.assert_expr(eq);
+            ctx.internalize_assertions();
+            val = fn;
+        }
         ctx.internalize(def, false);
         ctx.internalize(val.get(), false);
         return try_assign_eq(val.get(), def);
@@ -804,12 +818,30 @@ namespace smt {
         if (r == FC_DONE && m_bapa) {
             r = m_bapa->final_check();
         }
-        bool should_giveup = m_found_unsupported_op || has_propagate_up_trail();
+        bool should_giveup = m_found_unsupported_op || has_propagate_up_trail() || has_non_beta_as_array();
         if (r == FC_DONE && should_giveup)
             r = FC_GIVEUP;
         return r;
     }
 
+    bool theory_array_full::has_non_beta_as_array() {
+        for (enode* n : m_as_array) {
+            for (enode* p : n->get_parents())
+                if (!ctx.is_beta_redex(p, n)) {
+                    TRACE("array", tout << "not a beta redex " << enode_pp(p, ctx) << "\n");
+                    return true;
+                }
+        }
+        for (enode* n : m_lambdas) 
+            for (enode* p : n->get_parents())
+                if (!is_default(p) && !ctx.is_beta_redex(p, n)) {
+                    TRACE("array", tout << "lambda is not a beta redex " << enode_pp(p, ctx) << "\n");
+                    return true;
+                }
+        return false;
+    }
+
+
     bool theory_array_full::instantiate_parent_stores_default(theory_var v) {
         SASSERT(v != null_theory_var);
         TRACE("array", tout << "v" << v << "\n";);
